/* 
 * Linkoping Intelligent Communication of Knowledge System (LINCKS)
 *      Copyright (C) 1993, Lin Padgham, Ralph Rnnquist
 *       Department of Computer and Information Sciences
 *		University of Linkoping, Sweden
 *		    581 83 Linkoping, Sweden
 *		       lincks@ida.liu.se
 *
 * These collective LINCKS programs are free software; you can 
 * redistribute them and/or modify them under the terms of the GNU
 * General Public License as published by the Free Software Foundation,
 * version 2 of the License.
 *
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the programs; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* 
 * MODULE NAME: 	monitormain.c
 *
 * SCCSINFO:		@(#)monitormain.c	1.3 11/11/93
 *
 * ORIGINAL AUTHOR(S):  Peter {\AA}berg, 1987-01-20
 *
 * MODIFICATIONS:
 *	<list mods with name and date>
 *
 * DESCRIPTION:
 *
 *	This file contains the definition of main and the functions
 *	MakeName, Error, NewMolFile and NewDataFile
 *
 */

/*********************************************************************
 * INCLUDES:
 *********************************************************************/
#include <stdio.h>
#ifdef SCO
#include <sys/types.h>
#endif /* SCO */
#include <sys/file.h>
#ifndef SCO
#include <sys/types.h>
#endif /* n SCO */
#include <sys/wait.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>

#include "lincks.h"
#include "monitor.h"
#include "xconfig.h"
#include "libshared.h"
#include "systypes.h"
#include "config.h"

/*********************************************************************
 * EXTERNALLY-CALLABLE ROUTINES FOUND IN THIS MODULE:
 *********************************************************************/
/* none */

/*********************************************************************
 * EXTERNALLY-AVAILABLE	DATA FOUND IN THIS MODULE:
 *********************************************************************/
/* Global variables */
UID uidcom = SUPERUSER;		/* Currently processed user id */
int debug = 0;			/* Debugging flag */
int childpid;			/* PID of child i.e. net server */
char host[MAXNAMELEN];		/* Host name */
char lockfile[MAXPATHLEN];      /* lockfile name and path */

/* File descriptors and pointers */
int currentmolfd;		/* Currently open molecule file descriptor */
int currentdatafd;		/* Currently open data file descriptor */
int indexfd;			/* Index file descriptor */

/* File designators */
int currentdata;		/* Current data file number in table */
int currentmol;			/* Current molecule file number in table */

/* Data structures */
char *molnames[MAXNOOFFILES];	   /* Table of existing molecule file names */
char *datanames[MAXNOOFFILES];	   /* Table of existing data file names */
PROT *protlists[PROTLISTSIZE];	   /* Table of protection lists */
LOCK *locklist = NULL;		   /* Lock list tag */
INDEX indextable[INDEXTABLESIZE];  /* Table of index entries */
DBSQueue *penqueuep = NULL;        /* DBS queue */
LabelInfo *editinfop = NULL;       /* Edit information list */

/*********************************************************************
 * EXTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
/* labelinfo.c */
extern LabelInfo *CreateLabelInfo();

/* monerrors.c */
extern void Error(/* va_alist */);

/* rpccalls.c */
extern A_ENTRY *Access(/* L_ENTRY *args */);
extern A_ENTRY *Lock(/* L_ENTRY *args */);
extern N_ENTRY *Create(/* L_ENTRY *args */);
extern P_ENTRY *GetProt(/* L_ENTRY *args */);
extern C_ENTRY *GetCheckNo(/* L_ENTRY *args */);
extern C_ENTRY *UpdateCheckNo(/* L_ENTRY *args */);
extern int *Unlock(/* L_ENTRY *args */);
extern int *SetProt(/* S_ENTRY *args */);
extern int *Delete(/* L_ENTRY *args */);
extern int *Undelete(/* L_ENTRY *args */);
extern int *UpdateEntry(/* U_ENTRY *args */);
extern int *Kill(/* L_ENTRY *args */);
extern int *Restart(/* L_ENTRY *args */);
extern E_ENTRY *Check(/* R_ENTRY *args */);
extern E_ENTRY *Poll(/* int *pidp */);
extern int *Remove(/* R_ENTRY *args */);

/* monhandler.c */
extern RETSIGTYPE Handler(/* int sig, int code, struct sigcontext *scp */);

/*********************************************************************
 * EXTERNAL DATA STRUCTURES USED BY THIS MODULE:
 *********************************************************************/
/* none */

/*********************************************************************
 * LOCAL DEFINES, STRUCTS, TYPEDEFS, ETC.:
 *********************************************************************/
#define COMPILE_WARN \
"You must compile monitor with -DHAVE_SETPGRP2, -DHAVE_SETPGRP0\n\
or -DHAVE_SETSID or it will not work.\n"

/*********************************************************************
 * INTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
static int InitSignals();

/*********************************************************************
 * INTERNAL (STATIC) DATA: 
 *********************************************************************/
/* Database port number */
static u_long lincksport;		/* Database identity number */
static char *monnam;
static char netservfn[MAXPATHLEN]; /* Net server name */
static FILE *molnamesfp;	/* Molecule names file pointer */
static FILE *datanamesfp;	/* Data names file pointer */
static FILE *protectfp;		/* Protection file pointer */

/*  */
/**********************************************************************
 * Function: main(int argc, char **argv)
 *
 * Handles initialization of data structures and opening of
 * all system files. It also sets up the RPC communications
 * Arguments: [-d] debug mode
 *	dbid  database identity number
 * 
 * Modifications:
 *      <list mods with name and date>
 */
int main(argc, argv)
    int argc;
    char **argv;
{
    int c, errflag;
    int lockfd;
    char fname[MAXPATHLEN];
    int index, uid, grp;
    char filename[MAXPATHLEN];
    PROT *protlistp;
    FILE *fopen();
    char *malloc();
    extern int optind;
    extern DBSQueue *CreateDBSQueue();
    extern LabelInfo *CreateLabelInfo();
          
    /* RPC xdr definitions */
    int xdr_label(), xdr_access(), xdr_int(), xdr_setprot(), xdr_new();
    int xdr_prot(), xdr_checkno(), xdr_update(), xdr_u_short(), xdr_void();
    int xdr_edit(), xdr_remove(), xdr_int();

    /* RPC function definitions */
    A_ENTRY *Access(), *Lock();
    N_ENTRY *Create();
    P_ENTRY *GetProt();
    C_ENTRY *GetCheckNo(), *UpdateCheckNo();
    int *Unlock(), *SetProt(), *Delete(), *Undelete(), *UpdateEntry();
    int *Kill(), *Restart();
    E_ENTRY *Check(), *Poll();
    int *Remove();

    /* Get arguments from argv using getopt function */
    monnam = argv[0];
    errflag = 0;
    while ((c = getopt(argc, argv, "di:")) != EOF)
      switch(c) {
	case 'd':
	  debug = !debug;
	  break;
        default:
	  ++errflag;
	  break;
      }

    if (argc - optind == 1) {
      if (!configuration(argv[optind])) {
        /*
         * we have nowhere to "write" an error,
         * so this will have to do
         */
        exit(-1);
      }
      lincksport = (u_long) TCPIPNO;
    }
    else 
      ++errflag;

    /* If an illegal argument was given, signal error and exit */
    if (errflag) {
        (void)fprintf(stderr, "usage: monitor [-d] dbid\n");
	exit(1);
    }

    /* fork off in case we were started from command line - it 
     * should not be the case, but one never knows (this to get
     * our own process group) */
    switch( fork() )  {
    case -1 : /* failed to fork */
      perror("Failed to fork first child");
      exit(1);
    case 0 : /* child ... */
      break;
    default : /* parent */
      exit(0);
    }


#if !defined(HAVE_SETSID) && !defined(HAVE_SETPGRP0) && !defined(HAVE_SETPGRP2)
/*
 * You need to have one of HAVE_SETSID, HAVE_SETPGRP0 or HAVE_SETPGRP2 defined
 */
   (void)fprintf(stderr, COMPILE_WARN);
#endif

    /* Setup up process leader */
#ifdef HAVE_SETSID
    if ( setsid() < 0) {
#endif
#ifdef HAVE_SETPGRP0 /* System V */
    if ( setpgrp() == -1)  {
#endif
#ifdef HAVE_SETPGRP2 /* BSD */
    if ( setpgrp(0,getpid()) == -1)  {
#endif
      Error(ER_INIT,"Failed to change process group, errno %d", errno);
      exit(1);
    }

    /* Register program start */
    (void)gethostname(host, MAXNAMELEN);
    errno = 0;
    Error(0, "monitor %d started at %s", lincksport, host ? host : "");

    (void)sprintf(filename, "%s/%s", DBDIR, INDEXFILE);

    /* Open index file for read/write */
    if ((indexfd = open(filename, O_RDWR)) < 0) {
	Error(ER_INIT, 
	      "%s: could not open %s", monnam, filename );
	exit(1);
    }

    (void) sprintf(lockfile, "%s/%s", DBDIR, INDEXLOCK);
 
    /* this can fail under nfs with a loaded server, but ... */
    if ((lockfd = open(lockfile, O_WRONLY|O_CREAT|O_EXCL, 0644)) < 0) {
      Error(ER_INIT, "%s: could not create lockfile %s\n\
 Check if monitor or cutoff is already running.\n", monnam, lockfile);
      exit(1);
    }

    (void) sprintf( filename, "Monitor@%s:%d\n", host, getpid());
    if ( write( lockfd, filename, strlen(filename)) <= 0) {
      Error(ER_INIT, "%: failed to write to lockfile %s", monnam, lockfile);
      exit(1);
    }

    if ( close(lockfd) < 0 ) {
      Error(ER_INIT, "%s: failed to close lockfile %s", monnam, lockfile);
      exit(1);
    }

    /* Set up signals */
    if (InitSignals() < 0)
      goto quit;

    /* Create path to Net server */
    (void)sprintf(netservfn, "%s/%s", BINARIES, NETSERVER);

    /* Start Net Server process */
    errno = 0;
    Error(0, "netserv starting...");
    if ((childpid = fork()) == 0) {
        (void)Signal(SIGHUP, SIG_DFL);
	if (debug) {
	    if((execl(netservfn, NETSERVER, "-d", 
		      "-h", host, DBDIR, (char *)0)) == -1) {
	      Error(ER_INIT, "%s: execl failed to start %s",
		    monnam, netservfn);
	    }
	} else {
	    if((execl(netservfn, NETSERVER, "-h", host, 
		      DBDIR, (char *)0)) == -1) {
	      Error(ER_INIT, "%s: execl failed to start %s",
		    monnam, netservfn);
	    }
        }
	exit(1);
    }

    (void)sprintf(filename, "%s/%s", DBDIR, MOLNAMES);

    /* Open molecule names file for reading */
    if ((molnamesfp = fopen(filename, "r")) == NULL) {
	Error(ER_INIT, "%s: could not open %s", monnam, filename);
	goto quit;
    }
    index = 0;
    /* Read file contents into molecule names array */
    while (fscanf(molnamesfp, "%s\n", fname) != EOF) {
        if ((molnames[index] = malloc((unsigned)strlen(fname) + 1)) == NULL) {
	    Error(ER_INIT, "%s: no more memory for molecule names", monnam);
	    goto quit;
	}
        (void)strcpy(molnames[index++], fname);
    }
    currentmol = index - 1;
    (void)fclose(molnamesfp);

    (void)sprintf(filename, "%s/%s", DBDIR, molnames[currentmol]);

    /* Open current molecule file (the one added to) */
    if ((currentmolfd = open(filename, O_RDWR)) < 0) {
	Error(ER_INIT, "%s: could not open %s", monnam, filename);
	goto quit;
    }

    (void)sprintf(filename, "%s/%s", DBDIR, DATANAMES);

    /* Open data names file for reading */
    if ((datanamesfp = fopen(filename, "r")) == NULL) {
	Error(ER_INIT, "%s: could not open %s", monnam, filename);
	goto quit;
    }
    index = 0;
    /* Read file contents into data names array */
    while (fscanf(datanamesfp, "%s\n", fname) != EOF) {
        if ((datanames[index] = malloc((unsigned)strlen(fname) + 1)) == NULL) {
	    Error(ER_INIT, "%s: no more memory for data names", monnam);
	    goto quit;
	}
        (void)strcpy(datanames[index++], fname);
    }
    currentdata = index - 1;
    (void)fclose(datanamesfp);

    (void)sprintf(filename, "%s/%s", DBDIR, datanames[currentdata]);

    /* Open current data file (the one added to) */
    if ((currentdatafd = open(filename, O_RDWR)) < 0) {
	Error(ER_INIT, "%s: could not open %s", monnam, filename);
	goto quit;
    }

    (void)sprintf(filename, "%s/%s", DBDIR, PROTECTFILE);

    /* Open protection file for reading */
    if ((protectfp = fopen(filename, "r")) == NULL) {
	Error(ER_INIT, "%s: could not open %s", monnam, filename);
	goto quit;
    }

    /* Read file contents into protection lists */
    /* Lists are organized by protection group number */
    /* The file is organized by uid */
    while (fscanf(protectfp, "%d", &uid) != EOF) {
	while (fscanf(protectfp, "%d", &grp) == 1) {
	    if (protlists[grp] == NULL) {
		if ((protlists[grp] = (PROT *)malloc(PROTSIZE)) == NULL) {
		    Error(ER_INIT, "%s: no more memory for protlists",
			           monnam);
		    goto quit;
		}
		protlists[grp]->uid    = uid;
		protlists[grp]->p_next = NULL;
	    }
	    else {
		protlistp = protlists[grp];
		while (protlistp->p_next != NULL) 
		  protlistp = protlistp->p_next;
		if ((protlistp->p_next = (PROT *)malloc(PROTSIZE)) == NULL) {
		    Error(ER_INIT, "%s: no more memory for protlists",
			           monnam);
			 goto quit;
		     }
		protlistp->p_next->uid    = uid;
		protlistp->p_next->p_next = NULL;
	    }
	}
	if (getc(protectfp) == EOF)
	  break;
    }
    (void)fclose(protectfp);

    /* Init data structures */
    editinfop = (LabelInfo *) CreateLabelInfo();
    penqueuep = (DBSQueue *) CreateLabelInfo();
    
    /* Register the available RPC functions */
    if (
	registerrpc((u_long)MONITORPROGNO, lincksport, ACCESS, 
		    Access, xdr_label, xdr_access) ||
	registerrpc((u_long)MONITORPROGNO, lincksport, LOCKACCESS,
		    Lock, xdr_label, xdr_access) ||
	registerrpc((u_long)MONITORPROGNO, lincksport, UNLOCK,
		    Unlock, xdr_label, xdr_int) ||
	registerrpc((u_long)MONITORPROGNO, lincksport, SETPROT,
		    SetProt, xdr_setprot, xdr_int) ||
	registerrpc((u_long)MONITORPROGNO, lincksport, GETPROT,
		    GetProt, xdr_label, xdr_prot) ||
	registerrpc((u_long)MONITORPROGNO, lincksport, DELETE,
		    Delete, xdr_label, xdr_int) ||
	registerrpc((u_long)MONITORPROGNO, lincksport, UNDELETE,
		    Undelete, xdr_label, xdr_int) ||
	registerrpc((u_long)MONITORPROGNO, lincksport, GETCHECKNO,
		    GetCheckNo, xdr_label, xdr_checkno) ||
	registerrpc((u_long)MONITORPROGNO, lincksport, UPDATECHECKNO,
		    UpdateCheckNo, xdr_label, xdr_checkno) ||
	registerrpc((u_long)MONITORPROGNO, lincksport, UPDATEENTRY,
		    UpdateEntry, xdr_update, xdr_int) ||
	registerrpc((u_long)MONITORPROGNO, lincksport, CREATE,
		    Create, xdr_label, xdr_new) ||
	registerrpc((u_long)MONITORPROGNO, lincksport, KILL,
		    Kill, xdr_label, xdr_void) ||
	registerrpc((u_long)MONITORPROGNO, lincksport, RESTART,
		    Restart, xdr_label, xdr_void) ||
	registerrpc((u_long)MONITORPROGNO, lincksport, CHECK,
                    Check, xdr_remove, xdr_edit) ||
        registerrpc((u_long)MONITORPROGNO, lincksport, REMOVE,
                    Remove, xdr_remove, xdr_void) ||
        registerrpc((u_long)MONITORPROGNO, lincksport, POLL,
                    Poll, xdr_int, xdr_edit)
	) {
	/* registerrpc returned error status */
	Error(ER_INIT, "%s: registerrpc failure", monnam);
	goto quit;
    }

    /* Disconnect standard IO */
    (void)fflush(stderr);
    (void)close(fileno(stderr));
    (void)open("/dev/null", O_WRONLY);
    (void)fflush(stdout);
    (void)close(fileno(stdout));
    (void)open("/dev/null", O_WRONLY);
    (void)close(fileno(stdin));
    (void)open("/dev/null", O_WRONLY);

    /* Start RPC service - never returns */
    svc_run();

    /* If it does: note error and terminate system */
    Error(ER_INIT, "%s: svc_run returned", monnam);

 quit:
    /* Ignore child and kill (them) */
    (void)Signal(SIGCHLD, SIG_IGN);
    (void)kill(childpid, SIGHUP); 
    (void)wait((int *)0);
    exit(1);
#ifdef lint
/* lint is SOOOOOOOOOOOOO stupid sometimes */
    return 0;
#endif
}

/*  */
/**********************************************************************
 * Function: static int InitSignals()
 * 
 * Sets up monitor signals
 * SIGUSR1 starts a net server process
 *
 * Modifications:
 *      Sat Oct  9 19:28:05 1993 Martin Sjlin. Added check for SIGURG
 *                   and if not defined try SIGPOLL instead (SCO system).
 *      Thu Nov 11 10:25:19 1993 Martin Sjlin - SCO uses SIGUSR1 for
 *                   SIGURG (OOB) and SIGPOLL for async I/O.
 *      <list mods with name and date>
 */
static int InitSignals()
{
    /* Register signals to take care of */
    if (Signal(SIGALRM, Handler) == BADSIG) {
	Error(ER_INIT, "%s: could not register signal SIGALRM.", monnam);
	return(FAIL);
    }

    if (Signal(SIGCHLD, Handler) == BADSIG) {
	Error(ER_INIT, "%s: could not register signal SIGCHLD.", monnam);
	return(FAIL);
    }

#ifdef SIGXFSZ
    if (Signal(SIGXFSZ, Handler) == BADSIG) {
	Error(ER_INIT, "%s: could not register signal SIGXFSZ.", monnam);
	return(FAIL);
    }
#endif /* SIGXFSZ */
#ifndef SCO	
    if (Signal(SIGUSR1, Handler) == BADSIG) {
	Error(ER_INIT, "%s: could not register signal SIGUSR1.", monnam);
	return(FAIL);
    }
#endif /* n SCO */
	
    /* Register "clean-up" signals */
    if (Signal(SIGTERM, Handler) == BADSIG ) {
      Error(ER_INIT, "%s: could not register signal SIGTERM.", monnam);
      return(FAIL);
    }
    if (Signal(SIGHUP, Handler) == BADSIG ) {
      Error(ER_INIT, "%s: could not register signal SIGHUP.", monnam);
      return(FAIL);
    }
    if (Signal(SIGQUIT, Handler) == BADSIG ) {
      Error(ER_INIT, "%s: could not register signal SIGQUIT.", monnam);
      return(FAIL);
    }
#ifdef SIGLOST
    if (Signal(SIGLOST, Handler) == BADSIG ) {
      Error(ER_INIT, "%s: could not register signal SIGLOST.", monnam);
      return(FAIL);
    }
#endif

    /* Register signals to ignore */
    (void)Signal(SIGINT, SIG_IGN);
    (void)Signal(SIGPIPE, SIG_IGN);
#ifdef SIGURG
    (void)Signal(SIGURG, SIG_IGN);
#else
#ifdef SCO
    (void)Signal(SIGUSR1, SIG_IGN);
#else
    (void)Signal(SIGPOLL, SIG_IGN);
#endif /* n SCO */
#endif /* n SIGURG */
    return(SUCCESS);
}
