/* 
 * Linkoping Intelligent Communication of Knowledge System (LINCKS)
 *      Copyright (C) 1993, Lin Padgham, Ralph Rnnquist
 *       Department of Computer and Information Sciences
 *		University of Linkoping, Sweden
 *		    581 83 Linkoping, Sweden
 *		       lincks@ida.liu.se
 *
 * These collective LINCKS programs are free software; you can 
 * redistribute them and/or modify them under the terms of the GNU
 * General Public License as published by the Free Software Foundation,
 * version 2 of the License.
 *
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the programs; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * MODULE NAME:         liblincks.h
 *
 * SCCSINFO:            @(#)liblincks.h	1.1 9/27/93
 *
 * ORIGINAL AUTHOR(S):  Paul Svensson, 1988-11-23
 *
 * MODIFICATIONS:
 *	< add modifications with name and date >
 *
 * DESCRIPTION:
 * Declarations for use by applications that use the liblincks library.
 */

#ifndef LIBLINCKS_H
#define LIBLINCKS_H

typedef int errcode;
#define ERR_RCERROR	(-6)	/* DB: Error info sent              */
#define ERR_RCUUP	(-5)	/* DB: Unable to unpack molecule    */
#define ERR_RCBAD	(-4)	/* DB: Bad operation                */
#define ERR_ALREADY	(-3)	/* DB: Already received             */
#define ERR_STATUS	(-2)	/* DB: Unknown status returned      */
#define ERR_NODB	(-1)	/* Data base not connected	    */
#define NO_ERROR	  0	/* Normal return 		    */
#define ERR_LABEL	  1	/* Invalid label		    */
#define ERR_GROUP	  2	/* No such group		    */
#define ERR_TAG		  3	/* Too long tag			    */
#define ERR_FIELD	  4	/* No such field		    */
#define ERR_POS		  5	/* No such link item		    */
#define ERR_LOGIN	  6     /* Incorrect password or user       */
#define ERR_LASTLATEST	  7     /* Last item in latest              */
#define ERR_TH_LINK	  8	/* No TH_TRANSIENT last place link to node */
#define ERR_DOUBLE_EDITING 9	/* two transient versions created */
#define ERR_AMBIGUOUS 	  10	/* binding of unbound label ambiguous */
#define ERR_VERSION (11)
#define ERR_PARAMS	13	/* invalid parameters, likely null */
/* NOTE that dbcomm.c defines ERR_RESOLVE 12.... */

/* also in sunlincks.h, change both if you change here */
/* reserved group and field tags */
#define BTBL_GPD        "COMPOSITION","Gpd"
#define BTBL_LINK       "SYSTEM","Binding_Tables"
#define BTBL_PARTS      "COMPOSITION","Components"
#define BTBL_ROOT       "COMPOSITION","Root"
#define BUFFERNODE      "BUFFERNODE"
#define SYSTEM_CHAIN    "SYSTEM","Chain"
#define SYSTEM_COMMAND  "SYSTEM","Command"
#define SYSTEM_CREATED  "SYSTEM","Created"
#define SYSTEM_FORMER   "SYSTEM","Former_Latest"
#define SYSTEM_LATEST   "SYSTEM","Latest"
#define SYSTEM_OWNER    "SYSTEM","Owner"
#define SYSTEM_PARENT   "SYSTEM","Parent"
#define TH_TRANSIENT    "TH_TRANSIENT"
#define VSNET           "VSNET"
#define VSNET_COUNT     "VSNET","Count"

#define SUCCESS	  0    
#define FAIL	(-1)

/* GVS Protocol constants */
#define GVS_MORE	-1
#define GVS_INIT	0
#define GVS_VERSION	1
#define GVS_COMMAND	2
#define GVS_PREDECESSOR	3
#define GVS_CONTINUE	0
#define GVS_STOP	-1
#define GVS_NEXT	1

typedef int boolean;
#define FALSE	0
#define TRUE	1


typedef struct {
    int vs;		/* Version structure molecule index	*/
    			/* Identifier for a set of versions of a node. */
#define VS_SYS_ROOT	0	/* System root node. */

    int inst;		/* Instance molecule index		*/
			/* Identifier for a particular version of a node,
			 * or indicator for unbound reference. */
#define INST_UNBOUND	(-1)	/* Unbound link (old term: "free reference").*/
} label;

/*
 *	This record is for communicating attribute or image values
 *	with an application program.
 */
/*
 * NOTE: The following is also in sunlincks.h.
 * If you change this, also change sunlincks.h.
 */
typedef struct {
    int attsize;
    char *attvalue;
} attrval;

/* Pen message data structure definitions */
typedef struct penmsg {
  int index;	/* Index to object  */
  int type;	/* Message type     */
  char *msg;	/* Message          */
} PEN_MESSAGE;

/* this is provided by the library for logging errors */
extern void LogMess( /* varargs */ );

/*
 *	ANSI C Prototypes for the user functions found in usfuns.c
 */

#ifndef __PROTO__
#  ifdef __STDC__
#    define __PROTO__(ARGS) ARGS
#  else
#    define __PROTO__(ARGS) ()
#  endif
#endif

/* 
 * NOTE: if you change here, change in sunlincks.h as well.
 */
extern boolean ISTRANSIENT
    __PROTO__((label*));
extern boolean BOUND
    __PROTO__((label*));
extern errcode CAGN_CHANGEATTRGROUPNAME
    __PROTO__((label*, char*, char*));
extern errcode CAN_CHANGEATTRNAME
    __PROTO__((label*, char*, char*, char*));
extern errcode CLGN_CHANGELINKGROUPNAME
    __PROTO__((label*, char*, char*));
extern errcode CLN_CHANGELINKNAME
    __PROTO__((label*, char*, char*, char*));
extern errcode CO_CREATEOBJ
    __PROTO__((label*));
extern errcode CPO_COPYOBJ
    __PROTO__((label*, label*));
extern boolean EDITED
    __PROTO__((label*));
extern boolean OS_OBJECTSYNC
    __PROTO__((label*));
extern errcode GAGN_GETATTRGROUPNAMES
    __PROTO__((label*,
	       int(*)(char*, char*, int, int),
	       char*));
extern errcode GAN_GETATTRNAMES
    __PROTO__((label*,
	       char*,
	       int(*)(char*, char*, int, int),
	       char*));
extern errcode GA_GETATTR
    __PROTO__((label*, char*, char*, attrval*));
extern errcode GC_GETCURRENT
    __PROTO__((label*));
extern errcode GI_GETIMAGE
    __PROTO__((label*, attrval*));
extern errcode GLGN_GETLINKGROUPNAMES
    __PROTO__((label*,
	       int(*)(char*, char*, int, int),
	       char*));
extern errcode GLI_GETLINKITEM
    __PROTO__((label*, char*, char*, int, label*));
extern errcode GLN_GETLINKNAMES
    __PROTO__((label*,
	       char*,
	       int(*)(char*, char*, int, int),
	       char*));
extern errcode GLV_GETLINKVAL
    __PROTO__((label*,
	       char*,
	       char*,
	       int(*)(char*, label*, int, int),
	       char*));
extern errcode GL_GETLATEST
    __PROTO__((label*,
	       int(*)(char*, label*, int, int),
	       char*));
extern errcode GSR_GETSYSTEMROOT
    __PROTO__((label*));
extern errcode GVS_GETVERSIONS
    __PROTO__((label*,
	       int(*)(char*, label*, int, int),
	       char*));
extern errcode GBT_GETBINDINGTABLE
    __PROTO__((label *,label *,label *,int *));
extern errcode GBTC_GETBTCOMP
    __PROTO__((label *,label *,int, int));
extern boolean INWORKSPACE
    __PROTO__((label*));
extern errcode LINK
    __PROTO__((label*, char*, char*, label*, int));
extern errcode LOGIN
    __PROTO__((char*));
extern errcode NOPROMPT_LOGIN
    __PROTO__((char*, char*, char*));
extern errcode LOGOUT
    __PROTO__((void));
extern errcode PEN_RETRIEVE_MESSAGE
    __PROTO__((int(*)(char*, int, int, char*, int, int),
	      char*));	
extern errcode RAGT_REMOVEATTRGROUPTAG
    __PROTO__((label*, char*));
extern errcode RAGV_REMOVEATTRGROUPVAL
    __PROTO__((label*, char*));
extern errcode RAT_REMOVEATTRTAG
    __PROTO__((label*, char*, char*));
extern errcode RAV_REMOVEATTRVAL
    __PROTO__((label *, char *, char *));
extern errcode RFL_REMOVEFROMLATEST
    __PROTO__((label*));
extern errcode RLGT_REMOVELINKGROUPTAG
    __PROTO__((label*, char*));
extern errcode RLGV_REMOVELINKGROUPVAL
    __PROTO__((label*, char*));
extern errcode RLT_REMOVELINKTAG
    __PROTO__((label*, char*, char*));
extern errcode RLV_REMOVELINKVAL
    __PROTO__((label*, char*, char*));
extern errcode RO_RELEASEOBJ
    __PROTO__((label*));
extern errcode RTL_REMOVETRANSIENTLINKS
    __PROTO__((label*));
extern errcode SA_SETATTR
    __PROTO__((label*, char*, char*, attrval*));
extern errcode SI_SETIMAGE
    __PROTO__((label*, attrval*));
extern errcode SO_STOREOBJ
    __PROTO__((label*));
extern errcode UBL_UNBINDLABEL
    __PROTO__((label*));
extern errcode UL_UNLINK
    __PROTO__((label*, char*, char*, label*));
extern errcode ULA_UNLINKAT
    __PROTO__((label*, char*, char*, int));
#endif /* LIBLINCKS_H */
