/* 
 * Linkoping Intelligent Communication of Knowledge System (LINCKS)
 *      Copyright (C) 1993, Lin Padgham, Ralph Rnnquist
 *       Department of Computer and Information Sciences
 *		University of Linkoping, Sweden
 *		    581 83 Linkoping, Sweden
 *		       lincks@ida.liu.se
 *
 * These collective LINCKS programs are free software; you can 
 * redistribute them and/or modify them under the terms of the GNU
 * General Public License as published by the Free Software Foundation,
 * version 2 of the License.
 *
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the programs; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* 
 * MODULE NAME: 	lincks.h
 *
 * SCCSINFO:		@(#)lincks.h	1.1 9/27/93
 *
 * ORIGINAL AUTHOR(S):  Peter berg, Arne Stahre
 *
 * MODIFICATIONS:
 *	<list mods with name and date>
 *
 * DESCRIPTION:
 *	This is the Universal Lincks Include File. It contains essential
 *	data structure definitions and some system constants
 *	RPC procedure and program numbering is defined here
 */

#ifndef LINCKS_H
#define LINCKS_H

/* Global type definitions */
typedef int UID;
typedef unsigned char byte;

#ifndef FLAGS_DEFINED
typedef unsigned char FLAGS;
#define FLAGS_DEFINED
#endif /* FLAGS_DEFINED */

#ifndef LABEL_DEFINED
typedef long LABEL;
#define LABEL_DEFINED
#endif /* LABEL_DEFINED */

#ifndef SEEK_DEFINED
typedef long SEEK;
#define SEEK_DEFINED
#endif /* SEEK_DEFINED */

#define UNKNOWN	 ((UID) -1)
#define SUPERUSER ((UID)  0)

/* File entry size definition */
#define FILEMOLSIZE sizeof(FILEMOL)

/* Universal constants */
#define NOBLINKS  1	/* Signals that back-links should not be handled */

#ifndef MAXNAMELEN
#define MAXNAMELEN 48
#endif	/* MAXNAMELEN */

#ifndef MAXPATHLEN
#define MAXPATHLEN 512
#endif /* MAXPATHLEN */

#define MAXMESSAGELEN 32768
#define MAXOPENFILES 20
#define CHILDMINTIME 60
#ifndef NULL
#  define NULL 0
#endif
#define FAIL (-1)
#define SUCCESS 0

/* Monitor RPC procedure calls */
#define ACCESS ((unsigned long)1)
#define LOCKACCESS ((unsigned long)2)
#define UNLOCK ((unsigned long)3)
#define SETPROT ((unsigned long)4)
#define GETPROT ((unsigned long)5)
#define DELETE ((unsigned long)6)
#define UNDELETE ((unsigned long)7)
#define GETCHECKNO ((unsigned long)8)
#define UPDATECHECKNO ((unsigned long)9)
#define UPDATEENTRY ((unsigned long)10)
#define CREATE ((unsigned long)11)
#define KILL ((unsigned long)12)
#define RESTART ((unsigned long)13)
#define CHECK ((unsigned long)14)
#define REMOVE ((unsigned long)15)
#define POLL ((unsigned long)16)

/* Molecule file entry data structure definition */

/* File entry */
typedef struct {
	long flink_filepos;
	int flink_size;

	long blink_filepos;
	int blink_size;

	long attr_filepos;
	int attr_size;

	long image_filepos;
	int image_size;
} FILEMOL;

/*  */
/* XDR functions data definitions */

/* Access and lock */ 
typedef struct { 
	int a_status; 
	char *a_filename;
	long a_filepos;
	int a_checkno;
} A_ENTRY;

/* Getchecknum, updatechecknum */
typedef struct {
	int c_status;
	int c_checkno;
} C_ENTRY;

/* Getprot */
typedef struct {
	int p_status;
	unsigned short p_wrperm;
	unsigned short p_rdperm;
} P_ENTRY;

/* Createentry */
typedef struct {
	int n_status;
	LABEL n_label;
	char *n_filename;
	long n_filepos;
	int n_checkno;
} N_ENTRY;

/*
 * Parameters for: access, lock, unlock, getprot, delete, undelete
 * getcheckno, updatecheckno
 */
typedef struct {
	LABEL l_label;
	UID l_uid;
} L_ENTRY;

/* Parameters for: setprot */
typedef struct {
	LABEL s_label;
	UID s_uid;
	unsigned short s_wrperm;
	unsigned short s_rdperm;
} S_ENTRY;

/* Parameters for: updateentry */
typedef struct {
	LABEL u_label;
	UID u_uid;
	char *u_filename;
	long u_filepos;
} U_ENTRY;

/* Parameters for: remove, check */
typedef struct {
    LABEL r_label;
    int r_pid;
    char *r_username;
    int r_flag;
} R_ENTRY;

/* Edit info */
typedef struct {
    LABEL e_label;
    char *e_usernames;
} E_ENTRY;

/* Parallell editing notification */
typedef struct penmsg {
    int index;
    int type;
    char *msg;
} PEN_MESSAGE;

#endif /* LINCKS_H */
