/* 
 * Linkoping Intelligent Communication of Knowledge System (LINCKS)
 *      Copyright (C) 1993, Lin Padgham, Ralph Rnnquist
 *       Department of Computer and Information Sciences
 *		University of Linkoping, Sweden
 *		    581 83 Linkoping, Sweden
 *		       lincks@ida.liu.se
 *
 * These collective LINCKS programs are free software; you can 
 * redistribute them and/or modify them under the terms of the GNU
 * General Public License as published by the Free Software Foundation,
 * version 2 of the License.
 *
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the programs; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * MODULE NAME: 	aimstredit.c
 *
 * SCCSINFO:		@(#)aimstredit.c	1.3 12/15/93
 *
 * ORIGINAL AUTHOR(S):  1992-08-29, Ralph R\"onnquist
 *
 * DESCRIPTION:
 *	Structure editing code.
 *
 * MODIFICATIONS:
 *      1993-11-02 Martin Sjlin - whichref is moved to aimsubr.c and 
 *                 changed external section accordingly.
 *	<list mods with name and date>
 *
 */
/*********************************************************************
 * INCLUDES:
 *********************************************************************/
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#ifdef sgi
#include <X11/Xaw/Text.h>
#endif
#include <X11/Xaw/List.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include "aimtypes.h"
#include "aimstredit.h"

/*********************************************************************
 * EXTERNALLY-CALLABLE ROUTINES FOUND IN THIS MODULE:
 *********************************************************************/
void SHOWPOINTS();
void streditcmd( /* CALLBACK PROTOCOL */ );

/*********************************************************************
 * EXTERNALLY-AVAILABLE	DATA FOUND IN THIS MODULE:
 *********************************************************************/
/* none */

/*********************************************************************
 * EXTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
/* aimsubr.c */
extern reference_structure *whichref( /* infonode *in, identry *id,
					 int which */ );
/* aimcursor.c */
extern cursor templatecursor( /* cursor c */ );
extern cursor targetcursor( /* infonode *inp */ );
extern cursor freecursor( /* cursor c */ );
extern infonode *findinfonode( /* infonode *root,cursor c */ );
extern int sprintfcursor( /* char *buf,cursor c,int len */ );

/* aimcutbuf.c */
extern void copy_region( /* infonode *fst,infonode *lst,int cut */ );
extern void pop_cut_buffers( /* int dofree */ );

/* aimpaste.c */
extern int paste( /* infonode *inp, int after */ );

/* libc */
#ifdef sun
extern int printf();
#endif

/*********************************************************************
 * EXTERNAL DATA STRUCTURES USED BY THIS MODULE:
 *********************************************************************/
extern infonode *currentinfonode;		/* aimstart.c */
extern infonode *previousinfonode;		/* aimstruct.c */
extern Widget toplevel;			/* xstuff.c */

/*********************************************************************
 * LOCAL DEFINES, STRUCTS, TYPEDEFS, ETC.:
 *********************************************************************/
#if (XtSpecificationRelease < 5)
# ifndef sgi
/* This causes a recursive macro under IRIX (x11r4) */
#  define XawChainLeft XtChainLeft
#  define XawChainRight XtChainRight
#  define XawChainTop XtChainBottom
#  define XawChainBottom XtChainTop
# endif
#endif

/*********************************************************************
 * INTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/

/*********************************************************************
 * INTERNAL (STATIC) DATA:
 *********************************************************************/
static struct {
  Widget shell;
  Widget form;
  Widget closew;
  Widget list;
  Widget copyw;
  Widget cutw;
  Widget afterw;
  Widget beforew;
  Widget popw;
} menu = { (Widget)0, (Widget)0, (Widget)0, (Widget)0,
	   (Widget)0, (Widget)0, (Widget)0, (Widget)0,
	   (Widget)0};

static struct {
  char *text;
  int mode;
  } paste_modes[] = {
  {"Paste: share none, copy detailed", 2},
  {"Paste: share none, copy whole objects", 6},
  {"Paste: share leafs, copy detailed", 0},
  {"Paste: share leafs, copy whole objects", 4},
  {"Paste: share all, copy detailed", 8},
  {"Paste: share all, copy whole objects", 12}
};
static int paste_variant = 2;

static char *points[] =
{
  "[current  = ?]",
  "[previous = ?]",
  "Paste: share leafs, copy detailed",
  NULL
};

/*  */
/* ARGSUSED */
/**********************************************************************
 * Function: static void list_action(CALLBACK PROTOCOL)
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void list_action(w, client_data, call_data)
  Widget w;
  caddr_t client_data;
  caddr_t call_data;
{
  XawListReturnStruct *p = (XawListReturnStruct *)call_data;

  if (p->list_index == 2) {
    if (paste_variant++ >= 5)
      paste_variant = 0;
    points[2] = paste_modes[paste_variant].text;
    XawListChange(menu.list, points, 0, 0, True);
    return;
  }
  XawListUnhighlight(menu.list);
}

/* ^L */
/* ARGSUSED */
/**********************************************************************
 * Function: static void copy_button(CALLBACK PROTOCOL)
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void copy_button(w, client_data, call_data)
  Widget w;
  caddr_t client_data;
  caddr_t call_data;
{
  reference_structure *ref = (reference_structure *) 0;
  int cut = (int)client_data;

  if (!(ref = whichref(currentinfonode, (identry *)0, 0)) ||
      (ref != whichref(previousinfonode, (identry *)0, 0))) {
    (void)printf("Please mark two items within same view first.\n");
    return;
  }

  (void)printf("%sing region.\n",((cut)? "Cutt": "Copy"));
  copy_region(previousinfonode,currentinfonode,cut);
}

/* ^L */
/* ARGSUSED */
/**********************************************************************
 * Function: static void paste_button(CALLBACK PROTOCOL)
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void paste_button(w, client_data, call_data)
  Widget w;
  caddr_t client_data;
  caddr_t call_data;
{
  int after = (int)client_data;

  if (!currentinfonode) {
    (void)printf("Please mark an insertion point first.\n");
    return;
  }
  if (!(paste(currentinfonode,(paste_modes[paste_variant].mode|after))))
    XBell(XtDisplay(w), 50);
}

/* ^L */
/* ARGSUSED */
/**********************************************************************
 * Function: static void close_button(CALLBACK PROTOCOL)
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void close_button(w, client_data, call_data)
  Widget w;
  caddr_t client_data;
  caddr_t call_data;
{
  XtPopdown(menu.shell);
}

/*  */
/**********************************************************************
 * Function: void SHOWPOINTS()
 *
 * Modifications:
 *      <list mods with name and date>
 */
void SHOWPOINTS()
{
  static char curbuf[500];
  static char prebuf[500];
  cursor c;

  if ((currentinfonode == (infonode *) 0) ||
      ((c = targetcursor(currentinfonode)) == (cursor) 0)) {
    (void)strcpy(curbuf, "[current  = ?]");
  } else {
    (void)strcpy(curbuf, "[current  = ");
    (void)sprintfcursor(curbuf + 12, c, 500);
    (void)strcat(curbuf, " ]");
    (void)freecursor(c);
  }
  if ((previousinfonode == (infonode *) 0) ||
      ((c = targetcursor(previousinfonode)) == (cursor) 0)) {
    (void)strcpy(prebuf, "[previous  = ?]");
  } else {
    (void)strcpy(prebuf, "[previous = ");
    (void)sprintfcursor(prebuf + 12, c, 500);
    (void)strcat(prebuf, " ]");
    (void)freecursor(c);
  }

  points[0] = curbuf;
  points[1] = prebuf;
  if (menu.list == (Widget) 0)
    return;
  XawListChange(menu.list, points, 0, 0, True);
}

/*  */
/* ARGSUSED */
/*********************************************************************
 * Function: void streditcmd(CALLBACK PROTOCOL)
 *
 * Modification:
 *	<list mod's with name and date>
 */
void streditcmd(w, client_data, call_data)
  Widget w;
  caddr_t client_data;
  caddr_t call_data;
{
  SHOWPOINTS();

  if (menu.shell == (Widget) 0) {
    menu.shell = XtVaCreatePopupShell
      ("Structure Editing", applicationShellWidgetClass, toplevel,
       XtNallowShellResize, True,
       XtNborderWidth, 3,
       XtNminWidth, 240,
       XtNminHeight, 132,
       NULL);
    menu.form = XtVaCreateManagedWidget
      ("SED Form", formWidgetClass, menu.shell,
       XtNdefaultDistance, (XtArgVal) 8,
       NULL);
    menu.closew = XtVaCreateManagedWidget
      ("SED Copy", commandWidgetClass, menu.form,
       XtNlabel, "Close This Window",
       XtNwidth, (Dimension)240,
       XtNtop, XawChainTop,
       XtNbottom, XawChainTop,
       NULL);
    menu.list = XtVaCreateManagedWidget
      ("SED List", listWidgetClass, menu.form,
       XtNlist, points,
       XtNdefaultColumns, 1,
       XtNforceColumns, TRUE,
       XtNresizable, True,
       XtNwidth, (Dimension)240,
       XtNfromVert, menu.closew,
       XtNleft, XawChainLeft,
       XtNright, XawChainRight,
       XtNtop, XawChainTop,
       XtNbottom, XawChainTop,
       NULL);
    menu.copyw = XtVaCreateManagedWidget
      ("SED Copy", commandWidgetClass, menu.form,
       XtNlabel, "Copy Region",
       XtNwidth, (Dimension)100,
       XtNfromVert, menu.list,
       XtNleft, XawChainLeft,
       XtNright, XawChainLeft,
       XtNtop, XawChainTop,
       XtNbottom, XawChainTop,
       NULL);
    menu.cutw = XtVaCreateManagedWidget
      ("SED Cut", commandWidgetClass, menu.form,
       XtNlabel, "Cut Region",
       XtNwidth, (Dimension)100,
       XtNfromVert, menu.list,
       XtNfromHoriz, menu.copyw,
       XtNleft, XawChainLeft,
       XtNright, XawChainLeft,
       XtNtop, XawChainTop,
       XtNbottom, XawChainTop,
       NULL);
    menu.afterw = XtVaCreateManagedWidget
      ("SED Paste Before", commandWidgetClass, menu.form,
       XtNlabel, "Paste Before",
       XtNwidth, (Dimension)100,
       XtNfromVert, menu.copyw,
       XtNleft, XawChainLeft,
       XtNright, XawChainLeft,
       XtNtop, XawChainTop,
       XtNbottom, XawChainTop,
       NULL);
    menu.beforew = XtVaCreateManagedWidget
      ("SED Paste After", commandWidgetClass, menu.form,
       XtNlabel, "Paste After",
       XtNwidth, (Dimension)100,
       XtNfromVert, menu.copyw,
       XtNfromHoriz, menu.afterw,
       XtNleft, XawChainLeft,
       XtNright, XawChainLeft,
       XtNtop, XawChainTop,
       XtNbottom, XawChainTop,
       NULL);

    XtAddCallback(menu.closew, XtNcallback, close_button, (caddr_t)0);
    XtAddCallback(menu.copyw, XtNcallback, copy_button, (caddr_t)0);
    XtAddCallback(menu.cutw, XtNcallback, copy_button, (caddr_t)1);
    XtAddCallback(menu.afterw, XtNcallback, paste_button, (caddr_t)0);
    XtAddCallback(menu.beforew, XtNcallback, paste_button, (caddr_t)1);
    XtAddCallback(menu.list, XtNcallback, list_action, (caddr_t)0);
  }
  XtPopup(menu.shell, XtGrabNone);
}
