/* 
 * Linkoping Intelligent Communication of Knowledge System (LINCKS)
 *      Copyright (C) 1993, Lin Padgham, Ralph Rnnquist
 *       Department of Computer and Information Sciences
 *		University of Linkoping, Sweden
 *		    581 83 Linkoping, Sweden
 *		       lincks@ida.liu.se
 *
 * These collective LINCKS programs are free software; you can 
 * redistribute them and/or modify them under the terms of the GNU
 * General Public License as published by the Free Software Foundation,
 * version 2 of the License.
 *
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the programs; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* 
 * MODULE NAME: 	ltextaction.c
 *
 * SCCSINFO:		@(#)ltextaction.c	1.2 12/16/93
 *
 * ORIGINAL AUTHOR(S):  
 *
 * MODIFICATIONS:
 *	<list mods with name and date>
 *
 * DESCRIPTION:
 *	This file contains the key command action entries for the
 *	XlText widget.
 *
 */
/*********************************************************************
 * INCLUDES:
 *********************************************************************/
#include <stdio.h>
#include <ctype.h>
#include <X11/Xatom.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xmu/Atoms.h>
#include <X11/Xmu/Misc.h>
#include <X11/Xaw/TextP.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/Form.h>
#include "werror.h"
#include "ltextP.h"
#include "ltextsrcP.h"
#include "layout.h"
#include "view.h"	/* for XlScrollFocus to get widget class */
#include "systypes.h"

/*********************************************************************
 * EXTERNALLY-CALLABLE ROUTINES FOUND IN THIS MODULE:
 *********************************************************************/
void _XlTextCheckResize( /* TextWidget ctx */ );
XawTextPosition XlPositionForXY( /* TextWidget ctx, Position x,y */ );
void XltextMove(/* TextWidget ctx, XEvent *event, XawTextScanDirection dir,
	           XawTextScanType type, Boolean include */);
int XltextMult(/* TextWidget ctx,int mult */);
int XltextLastLineIndex( /* TextWidget w */ );
int XltextPoint( /* TextWidget ctx */ );
int XltextSearchForward( /* TextWidget ctx,XawTextPosition pos,char *text */ );
int XltextSearchBackward( /* TextWidget ctx,XawTextPosition pos,char *text */ );
void XltextSetFocus( /* Widget w, Position y */ );
int GivemeLine(/* TextWidget ctx, XawTextPosition position */);
void XltextPlaceCursor( /* TextWidget ctx,int line */ );
void TextLeave(/* Widget w, XEvent *event */);
int XltextYForCurrentPosition( /* TextWidget ctx */ );
void TextLeave();

/*********************************************************************
 * EXTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
/* aimcommand.c */
extern void closeallcmd();
extern void closecmd();
extern void closewidlcmd();
extern void dumpcmd();
extern void exportcmd();
extern void histcmd();
extern void nodehistcmd();
extern void promotecmd();
extern void quitcmd();
extern void refetchcmd();
extern void showgpdcmd();
extern void storecmd();
extern void store_all_cmd();

/* aimcutbuf.c */
extern void copy_region( /* infonode *fst,infonode *lst, int cut */ );
extern void make_cut_of_string( /* char *s */ );
extern int paste( /* infonode *inp, int after */ );

/* TextPop.c */
extern void _XawTextInsertFileAction();
extern void _XawTextInsertFile();
extern void _XawTextDoSearchAction();
extern void _XawTextDoReplaceAction();
extern void _XawTextSetField();
extern void _XawTextPopdownSearchAction();

/* Text.c */
extern char * _XawTextGetText();
extern void _XawTextBuildLineTable();
extern void _XawTextAlterSelection();
extern void _XawTextVScroll();
extern void _XawTextSetSelection();
extern void _XawTextCheckResize();
extern void _XawTextExecuteUpdate();
extern void _XawTextClearAndCenterDisplay();
extern Atom * _XawTextSelectionList();
extern _XawTextPrepareToUpdate(/* TextWidget ctx */);

/* generic.c */
extern void Xlgenvmove();
extern void Xlgenhmove();

/* xstuff.c */
extern void expandfn();
extern void popfn();
extern void selectfn();

/* ltextpop.c */
extern void XlTextAddDialogChildren(/* Widget form, Widget tw, char *ptr */);
extern void XlCreateChunkMenu(/* Widget w, XEvent *event */);
extern void XlCreateMiniBuffer(/* Widget w, XEvent *event, 
			       String *params, Cardinal *num_params */);
extern void XlTextCenterWidgetOnPoint(/* Widget w, XEvent *event */);
extern Widget XlTextCreateDialog(/* Widget parent, String ptr,
  				    String name, void (*func)() */);

/* ltextsink.c */
extern int CharWidth (/* Widget w, int x, unsigned char c */ );
extern int XlSinkMaxHeight( /* TextSinkObject w, int lines */ );

/* layout.c */
extern Widget XlChildIndex();
extern int XlHomeChildY( /* Widget w */ );
extern void LayoutFocusIn( /* Widget w */ );
extern void XlAddSearchChar(/* Widget w, char c */);
extern void XlEndSearch(/* Widget w, int flag */);
extern void XlDeleteSearchChar(/* Widget w */);
extern void XlBeginSearch(/* Widget w, int flag */);
extern void XlSetGoalColumn(/* Widget w, int c */);
extern void XlLayoutLostFocus( /* Widget w, int y */ );
extern void XlLayoutFocusDetails( /* Widget w, int y */ );
extern void XlSaveChildren( /* Widget w */ );
extern void ClearAllSelections( /* Widget w */ );

/* view.c */
extern void XlViewAutoScroll(/* Widget w,Position x,Position y,int frame */);
extern int XlViewPageScroll(/* Widget w, Boolean forward, Widget *nw,
			       Position *nx, Position *ny*/);
extern void XlViewCenterRow( /* Widget w, Position y */ );

/*********************************************************************
 * EXTERNAL DATA STRUCTURES USED BY THIS MODULE:
 *********************************************************************/
/* aimcommand.c */
extern infonode *previousinfonode;

/*********************************************************************
 * LOCAL DEFINES, STRUCTS, TYPEDEFS, ETC.:
 *********************************************************************/
typedef struct _SelectionList {
    String *params;
    Cardinal count;
    Time time;
} *SelectionList;

#ifdef DEBUG
/* since SyncHandle() requires that dpy be in local context */
#define sync(dpy) \
	if (dpy->synchandler) (*dpy->synchandler)(dpy); \
	sleep(1)
#endif /* DEBUG */

#define SrcScan                XawTextSourceScan
#define FindDist               XawTextSinkFindDistance
#define FindPos                XawTextSinkFindPosition

#define Point		ctx->text.insertPos
#define LastPos		ctx->text.lastPos
#define Source		ctx->text.source
#define Sink		ctx->text.sink
#define Mult		ctx->text.mult
#define LeftMargin	ctx->text.margin.left
#define RightMargin	ctx->text.margin.right
#define Info		ctx->text.lt.info
#define Lines		ctx->text.lt.lines

#define ReferenceP	ctx->xltext.reference
#define Inode		ReferenceP->inode
#define ChangeFlag	ctx->xltext.change_cnt

/*********************************************************************
 * INTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
static void AutoFill();
static void CloseAllWindows();
static void CloseWidls();
static void CloseWindow();
static void DeleteBackwardChar();
static void DeleteBackwardWord();
static void DeleteForwardWord();
static void DeleteBackwardWord();
static void _XawTextZapSelection();
static void DeleteForwardChar();
static void DeleteForwardWord();
static void DeleteOrKill();
static void DelPlural();
static void EndAction();
static void expand_in_new();
static void expand_in_place();
static void ExportWindow();
static void FormParagraph();
static void FormRegion();
static void GetSelection();
static void InsertChar();
static void InsertNewCRs();
static void InsertNewLine();
static void InsertNewLineAndBackup();
static int InsertNewLineAndBackupInternal();
static void InsertNewLineAndIndent();
static void InsertString();
static void InsPlural();
static void JoinText();
static void JoinTextInternal();
static void KillBackwardWord();
static void KillForwardWord();
static void KillToEndOfLine();
static void KillToEndOfParagraph();
static int LocalInsertNewLine();
static void Logout();
static void ModifySelection();
static void Move();
static void MoveBackwardChar();
static void MoveBackwardParagraph();
static void MoveBackwardWord();
static void MoveBeginningOfFile();
static void MoveEndOfFile();
static void MoveForwardChar();
static void MoveForwardParagraph();
static void MoveForwardWord();
static void MoveLine();
static void MoveNextLine();
static void MoveNextPage();
static void MovePage();
static void MovePreviousLine();
static void MovePreviousPage();
static void MoveToLineEnd();
static void MoveToLineStart();
static void NoOp();
static void pop_in_new();
static void pop_in_place();
static void PromoteVersion();
static void RedrawDisplay();
static void RefetchLatest();
static void SetAppend();

/* new stuff for X cut and paste for X buffers */
static void SelectStart();
static void ExtendAdjust();
static void ExtendEnd();
static void InsertSelection();
static void ExtendStart();
static void KillCurrentSelection();

static void ShowGpd();
static void ShowHistory();
static void ShowNodeHistory();
static void ShowRefStruct();
static void SplitText();
static void StartAction();
static void StoreAllWindows();
static void StoreWindow();
static void RaiseWin();
static void LowerWin();
static void IconifyWin();
static void TextFocusIn ();
static void TextFocusOut();
static void TextSearch();
static void TransposeCharacters();
void XlSavetoDB();
static void DoCommands();
static void XlScrollFocus();
static void _DeleteOrKill();
static void _SelectionReceived();
static int _XlLineForPosition();
static XawTextPosition StripOutOldCRs();

/*********************************************************************
 * INTERNAL (STATIC) DATA: 
 *********************************************************************/
static int global_appendflag = 0;
static int append_prev_state = 0;

/*********************************************************************
 * EXTERNALLY-AVAILABLE	DATA FOUND IN THIS MODULE:
 *********************************************************************/
int horrible_global_flag = 0;

/*
 * Handled only in ltextaction.c:EndAction and inspected in
 * ltextsink.c:ClearToBackground. Have a fit.
 */

/* Action Table */
XtActionsRec xltextActionsTable[] = {
/* motion bindings */
  {"forward-character", 	MoveForwardChar},
  {"backward-character", 	MoveBackwardChar},
  {"forward-word", 		MoveForwardWord},
  {"backward-word", 		MoveBackwardWord},
  {"forward-paragraph", 	MoveForwardParagraph},
  {"backward-paragraph", 	MoveBackwardParagraph},
  {"beginning-of-line", 	MoveToLineStart},
  {"end-of-line", 		MoveToLineEnd},
  {"next-line", 		MoveNextLine},
  {"previous-line", 		MovePreviousLine},
  {"next-page", 		MoveNextPage},
  {"previous-page", 		MovePreviousPage},
  {"beginning-of-file", 	MoveBeginningOfFile},
  {"end-of-file", 		MoveEndOfFile},
/* delete bindings */
  {"delete-next-character", 	DeleteForwardChar},
  {"delete-previous-character", DeleteBackwardChar},
  {"delete-next-word", 		DeleteForwardWord},
  {"delete-previous-word", 	DeleteBackwardWord},
/* kill bindings */
  {"kill-word", 		KillForwardWord},
  {"backward-kill-word", 	KillBackwardWord},
  {"join-text",			JoinText},
  {"split-text",		SplitText},
  {"kill-to-end-of-line", 	KillToEndOfLine},
  {"kill-to-end-of-paragraph", 	KillToEndOfParagraph},
/* new line stuff */
  {"newline-and-indent", 	InsertNewLineAndIndent},
  {"newline-and-backup", 	InsertNewLineAndBackup},
  {"newline", 			InsertNewLine},
/* expansion and popping stuff*/
  {"pop-in-new", 		pop_in_new},
  {"pop-in-place", 		pop_in_place},
  {"exp-in-new", 		expand_in_new},
  {"exp-in-place", 		expand_in_place},
/* Stuff to select parts of an item and put it in the X buffers */
  {"select-start",  		SelectStart},
  {"extend-adjust", 		ExtendAdjust},
  {"extend-end", 		ExtendEnd},
  {"insert-selection", 		InsertSelection},
  {"extend-start", 		ExtendStart},
  {"kill-selection",            KillCurrentSelection},
/* Save and Write bindings */
  {"save-text", 		XlSavetoDB},
  {"save-all-text", 		XlSaveChildren},
  {"do-commands", 		DoCommands}, 
  {"close-window",		CloseWindow},
  {"close-all-windows",		CloseAllWindows},
  {"store-window",		StoreWindow},
  {"store-all-windows",		StoreAllWindows},
  {"logout",			Logout},
  {"show-ref-struct",		ShowRefStruct},
  {"show-history",		ShowHistory},
  {"show-node-history",		ShowNodeHistory},
  {"show-gpd",			ShowGpd},
  {"export-window",		ExportWindow},
  {"close-widls",		CloseWidls},
/* Special Lincks commands which use meta-L */
  {"insert-plural",             InsPlural},
  {"delete-plural",             DelPlural},
  {"promote-version",		PromoteVersion},
  {"refetch-latest",		RefetchLatest},
/* Miscellaneous */
  {"redraw-display", 		RedrawDisplay},
  {"search",		        TextSearch},
  {"insert-char", 		InsertChar},
  {"insert-string",		InsertString},
  {"raise-win",			RaiseWin},
  {"lower-win",			LowerWin},
  {"iconify-win",		IconifyWin},
  {"focus-in", 	 	        TextFocusIn},
  {"focus-out", 		TextFocusOut},
  {"leave-window",		TextLeave},
  {"form-paragraph",            FormParagraph},
  {"transpose-characters",      TransposeCharacters},
  {"set-append",      		SetAppend},
  {"no-op",                     NoOp},
/* Action to bind special translations for text Dialogs. */
  {"InsertFileAction",          _XawTextInsertFileAction},
  {"DoSearchAction",            _XawTextDoSearchAction},
  {"DoReplaceAction",           _XawTextDoReplaceAction},
  {"SetField",                  _XawTextSetField},
  {"PopdownSearchAction",       _XawTextPopdownSearchAction},
};

Cardinal xltextActionsTableCount = XtNumber(xltextActionsTable);

/*  */
/**********************************************************************
  EDITING ACTIONS ARE DEFINED IN THE FOLLOWING CODE INCLUDES:
 **********************************************************************/
#include "ltextdeletion.c"
#include "ltextfocussing.c"
#include "ltextinsertion.c"
#include "ltextjustify.c"
#include "ltextmovement.c"
#include "ltextpublic.c"
#include "ltextsaving.c"
#include "ltextscrolling.c"
#include "ltextselection.c"
#include "ltextbuttons.c"


/*  */
/**********************************************************************
 * Function: static void StartAction(TextWidget ctx, XEvent *event)
 * 
 * Pick up event time and mouse position at event time.
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void StartAction(ctx, event)
  TextWidget ctx;
  XEvent *event;
{
  _XawTextPrepareToUpdate(ctx);

  if (event != NULL) {
    switch (event->type) {
    case ButtonPress:
    case ButtonRelease:
      ctx->text.time = event->xbutton.time;
      ctx->text.ev_x = event->xbutton.x;
      ctx->text.ev_y = event->xbutton.y;
      break;
    case KeyPress:
    case KeyRelease:
      ctx->text.time = event->xkey.time;
      ctx->text.ev_x = event->xkey.x;
      ctx->text.ev_y = event->xkey.y;
      break;
    case MotionNotify:
      ctx->text.time = event->xmotion.time;
      ctx->text.ev_x = event->xmotion.x;
      ctx->text.ev_y = event->xmotion.y;
      break;
    case EnterNotify:
    case LeaveNotify:
      ctx->text.time = event->xcrossing.time;
      ctx->text.ev_x = event->xcrossing.x;
      ctx->text.ev_y = event->xcrossing.y;
    }
  }
}


/*  */
/**********************************************************************
 * Function: static void EndAction(TextWidget ctx)
 * 
 * Modifications:
 *      <list mods with name and date>
 */
static void EndAction(ctx)
  TextWidget ctx;
{
  horrible_global_flag = 1;	/* Used in ltextsink.c:ClearToBackground */
  _XlTextCheckResize(ctx);
  _XawTextExecuteUpdate(ctx);
  Mult = 1;

  XlScrollFocus(ctx);

  XlSkipGoalSet((Widget)ctx,0);	/* Let goal column be set at next command. */
  XlEndSearch((Widget)ctx,0);		/* End searching as the normal case */
  horrible_global_flag = 0;	/* Used in ltextsink.c:ClearToBackground */
}


/*  */
/**********************************************************************
 * Function: static int _XlLineForPosition(TWO PARAMETERS)
 * Parameters:
 *	TextWidget ctx
 *	XawTextPosition position
 * 
 * This routine maps a source position in to the corresponding line number
 * of the text that is displayed in the window.
 *
 * This function depends on the existence of a valid line table!
 *
 * Modifications:
 *      <list mods with name and date>
 */
static int _XlLineForPosition (ctx, position)
  TextWidget ctx;
  XawTextPosition position;
{
  int line;
  
  for (line = 0; line < Lines; line++)
    if (position < Info[line + 1].position)
      break;
  return(line);
}


/*  */
/* ARGSUSED */
/**********************************************************************
 * Function: static void NoOp(FOUR PARAMETERS)
 * Parameters:
 *	Widget w		- text widget
 *	XEvent *event		- not used
 *	String *params		- action parameters
 *	Cardinal *num_params	- action parameters
 * 
 * This action performs no action, and allows the user or application
 * programmer to unbind a translation.
 *
 * Note: If the parameter list contains the string "RingBell" then
 *       this action will ring the bell.
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void NoOp(w, event, params, num_params)
  Widget w;
  XEvent *event;
  String *params;
  Cardinal *num_params;
{
    if (*num_params != 1)
	return;

    switch(params[0][0]) {
    case 'R':
    case 'r':
	XBell(XtDisplay(w), 0);
    default:			/* Fall Through */
	break;
    }
}

/*  */
/* ARGSUSED */
/**********************************************************************
 * Function: static void SetAppend(TRANSLATION PARAMETERS)
 * 
 * Modifications:
 *      <list mods with name and date>
 */
static void SetAppend(w, event, params, num_params)
  Widget w;
  XEvent *event;
  String *params;
  Cardinal *num_params;
{
  if (*num_params != 1)
    return;
  if ((append_prev_state == 1) && (params[0][0] == '1'))
    global_appendflag = 1;
  else
    global_appendflag = 0;
  append_prev_state = (params[0][0] == '0') ? 0 : 1;
}
