/* 
 * Linkoping Intelligent Communication of Knowledge System (LINCKS)
 *      Copyright (C) 1993, Lin Padgham, Ralph Rnnquist
 *       Department of Computer and Information Sciences
 *		University of Linkoping, Sweden
 *		    581 83 Linkoping, Sweden
 *		       lincks@ida.liu.se
 *
 * These collective LINCKS programs are free software; you can 
 * redistribute them and/or modify them under the terms of the GNU
 * General Public License as published by the Free Software Foundation,
 * version 2 of the License.
 *
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the programs; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * MODULE NAME: 	xdumpref.c
 *
 * SCCSINFO:		@(#)xdumpref.c	1.1 9/27/93
 *
 * ORIGINAL AUTHOR(S):  Ralph R\"onnquist,  Michael Jansson
 *
 * MODIFICATIONS:
 *      <list mods with name and date>
 *
 * DESCRIPTION:
 *	The operational code which produces reference structure dump.
 *
 */

/*********************************************************************
 * INCLUDES:
 *********************************************************************/
#include <stdio.h>
#ifndef sequent
#include <malloc.h>
#endif	/* sequent */

#include <X11/StringDefs.h>
#include <X11/Intrinsic.h>
#include <X11/Core.h>
#include <X11/Shell.h>
#include <X11/Composite.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Viewport.h>
#include <sys/param.h>		/* for MAXPATHLEN */
#include "aimtypes.h"
#include "layout.h"
#include "graphP.h"
#include "graphutils.h"
#include "systypes.h"

/*********************************************************************
 * EXTERNALLY-CALLABLE ROUTINES FOUND IN THIS MODULE:
 ********************************************************************/
Widget xdumpref( /* reference_structure *ref */ );
Widget dump_temporal_history( /* label *obj, reference_structure *r */ );
void whack_widl( /* CALLBACK PROTOCOL */ );

/*********************************************************************
 * EXTERNALLY-AVAILABLE	DATA FOUND IN THIS MODULE:
 *********************************************************************/
extern Widget toplevel;

/*********************************************************************
 * EXTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
/* aimbuildref.c */
extern reference_structure *buildref( /* label *tgobj, char *tag */ );
extern reference_structure *buildrefbt( /* char *tag, label *bt,
				int bound, reference_structure *oldref */ );
/* aimcmdstream.c */
extern void newlastcmd( /* int cmd, label *obj, int index, char *group,
                           char *field, int pos, attrval *val, label *lbl */ );

/* aimstruct.c */
extern void remove_widl( /* Widget widget, reference_structure *ref */);

/* aimsubr.c */
extern label *copylbl( /* label *tolbl, label *fromlbl */ );

/* showhistory.c */
extern int BuildHistory( /* GNode *root, label *lbl */ );
extern int BuildParentHistory( /* GNode *root, GNode *node, label *lbl */ );

/* graph.c */
extern void ClearGraphPart( /* GraphWidget req, GraphWidget graph */ );

/* graphutils.c */
extern void ComputeLayout( /* GraphNode *root, enum Gravity gravity,
		      	      int space_x, int space_y */ );
extern int GraphWidth( /* GraphNode *node */ );
extern int GraphHeight( /* GraphNode *node */ );
extern void FreeTree( /* GNode *root, void (*freefn)() */ );
extern GNode *CreateNode( /* GNode *parent, void *value */ );

/* rendering.c */
extern void ExtDumpGraph( /* GraphWidget graph, void (*dumpLink)(),
		             void (*dumpNode)(), void (*prolog)(),
			     void (*epilog)(), void (*comment)(),
			     void *context */ );
/* xstuff. c */
extern void fatal_error( /* va_alist */ );
extern void short_error( /* va_alist */ );
extern void popup_message( /* va_list */ );
extern XtArgVal CVfont( /* char *vp */ );

/* libshared */
extern char *strdup( /* char *incoming */ );

#ifdef sequent
extern char *malloc();
extern void free();
#endif  /* sequent */

/*********************************************************************
 * EXTERNAL DATA STRUCTURES USED BY THIS MODULE:
 *********************************************************************/
extern reference_structure *refstructs;    /* aimstart.c */

/*********************************************************************
 * LOCAL DEFINES, STRUCTS, TYPEDEFS, ETC.:
 *********************************************************************/
#ifndef MAXPATHLEN
#define MAXPATHLEN 512
#endif /* MAXPATHLEN */

#define ClearToBackground(w)	XClearWindow(XtDisplay(w),XtWindow(w))

typedef struct x_indentrec {	/* For the indent level stack */
  struct x_indentrec *next;
  int indent;
}

indentrec;
struct GraphCallData {
  GNode *template;
  GNode *target;
};
struct RefWidgets {
  GraphWidget target;
  GraphWidget template;
};
struct DumpCallData {
  GNode *root;
  reference_structure *ref;
};
struct MyValue {
  char *string;
  enum NodeType {
    identry_e, infonode_e
  } type;
};

#define Max(a, b) ((a) > (b) ? (a) : (b))

#define SPACEY 20
#define SPACEX 85
#define MARGIN 2
#define MAXWIDTH 700
#define MAXHEIGHT 500
#define MAXSTRLEN 10

/*
 *
 * This is the strings used when dumping an EPS file.
 * --------------------------------------------------
 */
#define A4HEIGHT 800
#define EPSDRAWNODE "%d %d %d (%d) Node\n"
#define EPSDRAWLINK "%d %d %d %d Link\n"
#define EPSDRAWINFONODE "%d %d (%s) DrawInfonode\n"
#define EPSDRAWIDENTRY "%d %d (%s) DrawIdentry\n"
#define EPSHEADER "%%!PS-Adobe-2.0 EPSF-1.2\n\
%%%%DocumentFonts:\n\
%%%%Pages: 1\n\
%%%%BoundingBox: 0 %d %d 900\n\
%%%%EndComments\n\
\n\
%% Subroutine for drawing a link between two nodes.\n\
/Link {\n\
newpath\n\
moveto\n\
lineto\n\
stroke\n\
} def\n\
\n\
%% Subroutine for drawing a node.\n\
/Node {\n\
/str exch def\n\
/r exch def\n\
/y exch def\n\
/x exch def\n\
newpath\n\
1 setgray\n\
x y r 0 360 arc\n\
fill\n\
0 setgray\n\
x y r 0 360 arc\n\
stroke\n\
/fontdict /Courier findfont 6 scalefont dup setfont\n\
x r 2 div sub y r 5 div sub moveto\n\
str show\n\
fill\n\
} def\n\
%% Subroutine for drawing an infonode.\n\
/DrawInfonode {\n\
/str exch def\n\
/y exch def\n\
/x exch def\n\
x y 2 sub moveto\n\
x 60 add y 2 sub lineto\n\
x 60 add y 8 add lineto\n\
x y 8 add lineto\n\
x y 2 sub lineto\n\
closepath\n\
gsave\n\
1 setgray\n\
fill\n\
grestore\n\
stroke\n\
0 setgray\n\
/fontdict /Courier findfont 9 scalefont dup setfont\n\
x 2 add y moveto\n\
str show\n\
fill\n\
} def\n\
\n\
%% Subroutine for drawing an identry.\n\
/DrawIdentry {\n\
/str exch def\n\
/y exch def\n\
/x exch def\n\
0 setgray\n\
/fontdict /Courier findfont 9 scalefont dup setfont\n\
x 2 add y 2 add moveto\n\
str show\n\
fill\n\
} def\n\
\n\
%%%%EndProlog"

/*********************************************************************
 * INTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
/* Makes window title from ref.struct */
static char *makeshelltitle( /* reference_structure *ref */ );

/* Makes access string from infonode */
static void makeaccessline( /* infonode *inp,char *buf */ );

/* Makes contents string from infonode */
static void makevalueline( /* infonode *inp,char *buf */ );

/* Prints the value of identry */
static int stepinfonodes( /* infonode *inp, int row, int indent,
		   	     indentrec *lst, GNode *parent */ );
/* Prints the subnodes of infonode */
static int stepidentries( /* identry *idp, int row, int indent,
		 	     indentrec *lst, GNode *parent */ );

static void DrawRefNode( /* GraphWidget graph, GC gc, int x, int y,
	         	    int size, struct MyValue *value */ );
static void closeref( /* Widget  widget, XtPointer client_data,
		      	 XtPointer call_data */ );
static void closehist( /* Widget  widget, XtPointer client_data,
		       	  XtPointer call_data */ );
static void EPSdumpLink(/* GraphWidget graph, GC gc, int x1, int y1,
		      	   int x2, int y2, int size */ );
static void EPSDumpRefNode( /* GraphWidget graph, GC gc, int x, int y,
	       		       int size, struct MyValue *value */ );
static void EPSProlog( /* void *context, int width, int height */ );
static void EPSEpilog( /* void *context */ );
static void EPSComment( /* void *context, char *str */ );
static void DumpRef( /* Widget  widget, XtPointer client_data,
		     	XtPointer call_data */ );
static void DumpHistory( /* Widget widget, XtPointer client_data,
			    XtPointer call_data */ );
static void ack( /* GraphWidget widget, GNode *node */ );
static void destroywid( /* CALLBACK PROTOCOL */ );
static void whack_refs( /* CALLBACK PROTOCOL */ );
static void freeMyValue( /* GNode *node */ );
static void mark_graph( /* GNode *root */ );
static void drawlink( /* GraphWidget graph, GC gc, int x1, int y1,
			 int x2, int y2, int size */ );

/*********************************************************************
 * INTERNAL (STATIC) DATA:
 *********************************************************************/
static char buf[1000];		/* Scratch pad */

/*  */
/***********************************************************************
 * Function: static char *makeshelltitle(reference_structure *ref)
 *
 * Constructs a descriptive name (into buf) for the given reference
 * structure. Returns buf.
 * NOTE: this is NOT a copy of the function in xstuff.c with same name.
 *
 * The name is formed from both the given reference structure and the
 * GPD used. It will look like the following:
 *	"A target identific] using [View of text"
 * The target identification is formed similar to how window titles are
 * formed (see xstuff.c), and the GPD identification is obtained from
 * the image of the GPD (root) object.
 * Modifications:
 *	<list mods here with name and date>
 */
static char *makeshelltitle(ref)
  reference_structure *ref;
{
  identry *idp = ref->target;
  attrval v;

  /* Make a target identification string */
  if (!(idp && idp->value && idp->value->val.attsize)) {
    (void)strcat(buf, "(no name)");
  } else {
    if (idp->value->val.attsize > 25) {
      (void)strncpy(buf, idp->value->val.attvalue, 25);
      (void)strcpy(&buf[25], " ...");
    } else {
      (void)strcpy(buf, idp->value->val.attvalue);
    }
  }

  /* Make a GPD identification string */
  if (GI_GETIMAGE(&(ref->gpd), &v) != SUCCESS) {
    v.attvalue = strdup("(view not described)");
    v.attsize = strlen(v.attvalue) + 1;
  }
  if ((int)strlen(v.attvalue) > 25)
    v.attvalue[25] = '\0';

  /* Combine target and GPD identifications into shell title string */
  (void)strcat(buf, "] using [");
  (void)strcat(buf, v.attvalue);
  if (v.attvalue)
    free(v.attvalue);
  return buf;
}

/*  */
/**********************************************************************
 * Function: static void makeaccessline(infonode *inp,char *buf);
 *
 * Puts an access string into buf for a template(!!) infonode.
 * Returns nothing.
 *
 * The access string looks in principle the same as it does in the
 * GPD.
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void makeaccessline(inp, buf)
  infonode *inp;
  char *buf;
{
  switch (inp->index) {
  case INTERNAL:
    buf[0] = '"';
    (void)strcpy(buf + 1, inp->val.attvalue);
    (void)strcat(buf, "\"");
    break;
  case ATTRVALUE:
  case ATTRFIELDTAG:
  case ATTRGROUPTAG:
    (void)strcpy(buf, "ATTR ");
    if (inp->group) {
      (void)strcat(buf, inp->group);
      if (inp->field) {
	(void)strcat(buf, " ");
	(void)strcat(buf, inp->field);
      }
    }
    break;
  case LINKITEM:
  case LINKFIELDTAG:
  case LINKGROUPTAG:
    (void)strcpy(buf, "LINK ");
    if (inp->group) {
      (void)strcat(buf, inp->group);
      if (inp->field) {
	(void)strcat(buf, " ");
	(void)strcat(buf, inp->field);
	if (inp->pos != -1) {
	  (void)strcat(buf, " 1");
	}
      }
    }
    break;
  case IMAGE:
    (void)strcpy(buf, "IMAGE");
    break;
  case SAMEASPARENT:
    (void)strcpy(buf, "$");
    break;
  default:
    (void)strcpy(buf, "(null)");
    break;
  }
}

/*  */
/**********************************************************************
 * Function: static void makevalueline(infonode *inp,char *buf)
 *
 * Puts an infonode value string into buf.
 * Returns nothing.
 *
 * The infonode value string consists of up to MAXSTRLEN characters from the
 * val.attvalue string, or the constant "(null)" as appropriate.
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void makevalueline(inp, buf)
  infonode *inp;
  char *buf;
{
  int i;

  if (inp->val.attvalue == (char *)0) {
    (void)strcpy(buf, "(null)");
    return;
  }
  for (i = 0; (buf[i] = (inp->val.attvalue)[i]) != '\0'; i++)
    if (i == MAXSTRLEN)
      break;
  buf[i] = '\0';
  if (i == 0)
    (void)strcpy(buf, "(null)");
}

/*  */
/**********************************************************************
 * Function: static int stepinfonodes(FIVE PARAMETERS)
 * Parameters:
 *	infonode *inp
 *	int row
 *	int indent
 *	indentrec *lst
 *	GNode *parent
 *
 * The sub-tree is printed to the right of 'indent' column, changing row
 * as needed, and using the indentation stack 'lst' for indentation prefix.
 *
 * Infonodes are printed using the infonode value string produced by
 * makeaccessline (for template) or makevalueline (for target), within
 * square brackets, and preceded by the constant string "---".
 *
 * Modifications:
 *      <list mods with name and date>
 */
static int stepinfonodes(inp, row, indent, lst, parent)
  infonode *inp;
  int row;
  int indent;
  indentrec *lst;
  GNode *parent;
{
  indentrec *p;
  struct MyValue *val = NULL;
  int x, y, width;

  if (inp == (infonode *)0)
    return (0);

  if (inp->head->tmpl == (identry *)0) {
    /* Dumping the template; display is the access sentence */
    makeaccessline(inp, buf);
  } else {
    /* Dumping the target; display portion of value */
    makevalueline(inp, buf);
  }

  if (strlen(buf)) {
    val = (struct MyValue *)malloc(sizeof(struct MyValue));
    if (val == NULL) {
      fatal_error("%s\n", "malloc failed in stepinfonodes");
    }
    val->string = (char *)malloc((unsigned)strlen(buf) + 1);
    if (val->string == NULL) {
      fatal_error("%s\n", "malloc failed in stepinfonodes");
    }
    width = ((int)strlen(buf) + 
	     (int)strlen(((struct MyValue *)(parent->node.value))->string))/2;
    x = parent->node.x + (width * 10);
    y = parent->node.y;
    (void)strcpy(val->string, buf);
    val->type = infonode_e;
    parent = CreateNode(parent, (void *)val);
    parent->node.x = x;
    parent->node.y = y;
  }
  if ((p = (indentrec *) malloc(sizeof(indentrec))) == NULL) {
    fatal_error("%s\n", "malloc failed in stepinfonodes");
  }
  p->indent = indent + 4;
  p->next = lst;
  if (inp->next != (infonode *)0) {
    row = stepinfonodes(inp->next, row, indent + strlen(buf) + 4, p, parent);
  }
  row = stepidentries(inp->subnodes, row, indent + 4, p, parent);
  if (inp->next == NULL)
    row += SPACEY / 2;
  free((char *)p);
  return row;
}

/*  */
/**********************************************************************
 * Function: static int stepidentries(FIVE PARAMETERS)
 * Parameters:
 *	identry *idp
 * 	int row
 *	int indent
 *	indentrec *lst
 *	GNode *parent
 *
 * Identries are printed using the indentation of makeindent for given
 * 'indent' and 'lst', followed by idp->idname.
 *
 * Modifications:
 *      <list mods with name and date>
 */
static int stepidentries(idp, row, indent, lst, parent)
  identry *idp;
  int row;
  int indent;
  indentrec *lst;
  GNode *parent;
{
  struct MyValue *val = NULL;
  int x;

  for (; idp; idp = idp->next) {
    if (idp->idname && strlen(idp->idname) &&
	(val = (struct MyValue *)malloc(sizeof(struct MyValue))) &&
	  (val->string = (char *)malloc((unsigned)strlen(idp->idname) + 1))) {
      (void)strcpy(val->string, idp->idname);
      val->type = identry_e;
      x = parent->node.x;
      row += SPACEY;
      parent = CreateNode(parent, (void *)val);
      parent->node.x = x;
      parent->node.y = row;
    }
    if (idp->value->head == idp) {
      if ((lst == (indentrec *) 0) || (idp->next != (identry *)0)) {
	row = stepinfonodes(idp->value, row, indent + strlen(idp->idname),
			    lst, parent);
      } else {
	row = stepinfonodes(idp->value, row, indent + strlen(idp->idname),
			    lst->next, parent);
      }
    }
  }
  return row;
}

/*  */
/*ARGSUSED*/
/**********************************************************************
 * Function: static void DrawRefNode(SIX PARAMETERS)
 * Parameters:
 *	GraphWidget graph
 *	GC gc
 *	int x
 *	int y
 *	int size
 *	struct MyValue *value
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void DrawRefNode(graph, gc, x, y, size, value)
  GraphWidget graph;
  GC gc;
  int x;
  int y;
  int size;
  struct MyValue *value;
{
  int width;
  int height;
  struct {
    int x;
    int y;
    unsigned int width;
    unsigned int height;
 } r;

  /* Calculate the dimensions. */
  if (graph->graph.font) {
    width =
      XTextWidth(graph->graph.font, value->string, strlen(value->string));
    height = graph->graph.font->ascent;
    x -= width / 2;
    y += height / 2;
  }
  r.x = x - MARGIN;
  r.y = y - MARGIN - height;
  r.width = (unsigned int)width + 2 * MARGIN;
  r.height = (unsigned int)height + 2 * MARGIN;

  /* Clear the background of the node. */
  XSetForeground(XtDisplay(graph), gc, WhitePixelOfScreen(XtScreen(graph)));

  XFillRectangle(XtDisplay(graph), XtWindow(graph), gc,
		 r.x, r.y, r.width, r.height);

  XSetForeground(XtDisplay(graph), gc, BlackPixelOfScreen(XtScreen(graph)));

  /* Draw a border. */
  if (value->type == infonode_e)
    XDrawRectangle(XtDisplay(graph), XtWindow(graph), gc,
		   r.x, r.y, r.width, r.height);

  /* Print the value of the node in the middle of the rectangle. */
  XDrawString(XtDisplay(graph), XtWindow(graph), gc,
	      x, y, value->string, strlen(value->string));
}

/*  */
/* ARGSUSED */
/**********************************************************************
 * Function: static void whack_refs(CALLBACK PROTOCOL)
 * 
 * free a RefWidgets structure 
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void whack_refs(widget, client_data, call_data)
  Widget widget;
  XtPointer client_data, call_data;
{
  struct RefWidgets *refs = (struct RefWidgets *)client_data;
  free((char *)refs);
}

/*  */
/* ARGSUSED */
/**********************************************************************
 * Function: void whack_widl(CALLBACK PROTOCOL)
 * 
 * when a graph window is closed, its corresponding widl must be removed
 * from the reference structure widl list.
 *
 * Modifications:
 *      <list mods with name and date>
 */
void whack_widl(widget, client_data, call_data)
  Widget widget;
  XtPointer client_data, call_data;
{
  reference_structure *widref = (reference_structure *)client_data;
  reference_structure *ref = NULL;
  
  for (ref = refstructs; ref; ref = ref->next)
    if (ref == widref)
      break;

  if (ref)
    remove_widl(widget, widref);
}

/*  */
/*ARGSUSED*/
/**********************************************************************
 * Function: static void closeref(CALLBACK PROTOCOL)
 *
 * Frees the two graphs and closes the window associated with a
 * references structure dump.
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void closeref(widget, client_data, call_data)
  Widget widget;
  XtPointer client_data, call_data;
{
  struct GraphCallData *graph = (struct GraphCallData *)client_data;

  if (!graph) {
    (void)fprintf(stderr, "closeref: (warning) null graph input param.\n");
    return;
  }

  FreeTree(graph->template, freeMyValue);
  FreeTree(graph->target, freeMyValue);
  free((char *)(graph));
}

/*  */
/*ARGSUSED*/
/**********************************************************************
 * Function: static void destroywid(THREE PARAMETERS)
 * Parameters:
 *	Widget  widget
 *	XtPointer client_data
 *	XtPointer call_data
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void destroywid(widget, client_data, call_data)
  Widget widget;
  XtPointer client_data, call_data;
{
  Widget shell = (Widget)client_data;
  XtDestroyWidget(shell);
}


/*  */
/*ARGSUSED*/
/**********************************************************************
 * Function: static void closehist(THREE PARAMETERS)
 * Parameters:
 *	Widget  widget
 *	XtPointer client_data
 *	XtPointer call_data
 *
 * Free the graph of the (GraphWidget)client_data.
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void closehist(widget, client_data, call_data)
  Widget widget;
  XtPointer client_data, call_data;
{
  GraphWidget graph = (GraphWidget)client_data;

  if (!graph) {
    (void)fprintf(stderr, "closeref: (warning) null graph input param.\n");
    return;
  }
  if (!graph->graph.root)
    return;
  mark_graph(graph->graph.root);
  FreeTree(graph->graph.root,NULL);
}

/*  */
/**********************************************************************
 * Function: static void mark_graph(GNode *root)
 * 
 * Go through the graph and make it into a tree for freeing purposes.
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void mark_graph(root)
  GNode *root;
{
  int i;

  root->num = -936;
  for (i=0; i<root->used; i++)
    if (((GNode *)root->node.children[i])->num == -936)
      root->node.children[i] = NULL;
    else
      mark_graph((GNode *)(root->node.children[i]));
}

/*  */
/*
 * The following functions are callback functions used in the
 * ExtDumpGraph()/render.c function.
 */

/*ARGSUSED*/
/**********************************************************************
 * Function: static void EPSdumpNode(SIX PARAMETERS)
 * Parameters:
 *	GraphWidget graph;
 *	GC gc;
 *	int x, y, size;
 *	void *value;
 * 
 * Modifications:
 *      <list mods with name and date>
 */
static void EPSdumpNode(graph, gc, x, y, size, value)
  GraphWidget graph;
  GC gc;
  int x, y, size;
  void *value;
{
  (void)fprintf((FILE *) gc, EPSDRAWNODE, x, A4HEIGHT - y, size / 2,
		(int)value);
}

/*  */
/*ARGSUSED*/
/**********************************************************************
 * Function: static void EPSdumpLink(SEVEN PARAMETERS)
 * Parameters:
 *	GraphWidget graph
 *	GC gc
 *	int x1, y1, x2, y2, size
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void EPSdumpLink(graph, gc, x1, y1, x2, y2, size)
  GraphWidget graph;
  GC gc;
  int x1, y1, x2, y2, size;
{
  (void)fprintf((FILE *) gc, EPSDRAWLINK, x1, A4HEIGHT - y1, x2, 
		A4HEIGHT - y2);
}

/*  */
/*ARGSUSED*/
/**********************************************************************
 * Function: static void EPSDumpRefNode(SIX PARAMETERS)
 * Parameters:
 *	GraphWidget graph
 *	GC gc
 *	int x, y, size
 *	struct MyValue *value
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void EPSDumpRefNode(graph, gc, x, y, size, value)
  GraphWidget graph;
  GC gc;
  int x, y, size;
  struct MyValue *value;
{
  char string[MAXPATHLEN];
  int i, j;

  /* Quote all "(" and ")" and "\". */
  for (i = 0, j = 0; i < (int)strlen(value->string); i++) {
    if ((value->string[i] == '(')
	|| (value->string[i] == ')')
	|| (value->string[i] == '\\'))
      string[j++] = '\\';
    string[j++] = value->string[i];
  }
  string[j] = '\0';

  if (value->type == infonode_e)
    (void)fprintf((FILE *) gc, EPSDRAWINFONODE, x, A4HEIGHT - y, string);
  else
    (void)fprintf((FILE *) gc, EPSDRAWIDENTRY, x, A4HEIGHT - y, string);
}

/*  */
/**********************************************************************
 * Function: static void EPSProlog(void *context, int width, int height)
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void EPSProlog(context, width, height)
  void *context;
  int width;
  int height;
{
  (void)fprintf((FILE *) context, EPSHEADER, width, A4HEIGHT - height);
}

/*  */
/**********************************************************************
 * Function: static void EPSEpilog(void *context)
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void EPSEpilog(context)
  void *context;
{
  (void)fputs("\nshowpage\n\nend\n", (FILE *) context);
}

/*  */
/**********************************************************************
 * Function: static void EPSComment(void *context, char *str)
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void EPSComment(context, str)
  void *context;
  char *str;
{
  (void)fprintf((FILE *) context, "\n\n%% %s\n", str);
}

/*  */
/*ARGSUSED*/
/**********************************************************************
 * Function: static void DumpRef(THREE PARAMETERS)
 * Parameters:
 *	Widget  widget
 *	XtPointer client_data, call_data
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void DumpRef(widget, client_data, call_data)
  Widget widget;
  XtPointer client_data, call_data;
{
  FILE *file;
  char filename[MAXPATHLEN];
  char stemname[MAXPATHLEN];
  struct RefWidgets *refs = (struct RefWidgets *)client_data;

  /* Open a file */
  (void)printf("Specify the file to write to: ");
  (void)scanf("%s", stemname);

  /* Strip extension of the name. -- unnecessary! if they give a name,
  use it! */
  /*
  if (strchr(stemname, '.'))
    strchr(stemname, '.')[0] = '\0';
  */

  /* Dump template! */
  (void)strcpy(filename, stemname);
  (void)strcat(filename, "_tmpl.eps");
  (void)printf("Writing template to %s.\n", filename);

  if (!(file = fopen(filename, "w"))) {
    (void)fputs("Failed to open that file. No action taken!", stderr);
    return;
  }
  ExtDumpGraph(refs->template,
	       EPSdumpLink,
	       EPSDumpRefNode,
	       EPSProlog,
	       EPSEpilog,
	       EPSComment,
	       (void *)file);
  if ((fclose(file)) != 0) {
    (void)fprintf(stderr, "fclose failed in DumpRef.\n");
  }
  /* Dump target! */
  (void)strcpy(filename, stemname);
  (void)strcat(filename, "_trgt.eps");
  (void)printf("Writing target in %s.\n", filename);

  if (!(file = fopen(filename, "w"))) {
    (void)fputs("Failed to open that file. No action taken!", stderr);
    return;
  }
  ExtDumpGraph(refs->target,
	       EPSdumpLink,
	       EPSDumpRefNode,
	       EPSProlog,
	       EPSEpilog,
	       EPSComment,
	       (void *)file);

  if ((fclose(file)) != 0) {
    (void)fprintf(stderr, "fclose failed in DumpRef.\n");
  }
  (void)printf("Done.\n");
}

/*  */
/*ARGSUSED*/
/**********************************************************************
 * Function: static void DumpHistory(THREE PARAMETERS)
 * Parameters:
 *	Widget  widget
 *	XtPointer client_data, call_data
 *
 * Load the version structure and generate a postscript file with the graph
 * of the history. Works more or less like the GraphWidget.
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void DumpHistory(widget, client_data, call_data)
  Widget widget;
  XtPointer client_data, call_data;
{
  FILE *file;
  char filename[MAXPATHLEN];

  /* Open a file */
  (void)printf("Specify the file to write to: ");
  (void)scanf("%s", filename);
  if (!(file = fopen(filename, "w"))) {
    (void)fputs("Failed to open that file. No action taken!", stderr);
    return;
  }
  /* Dump it! */
  ExtDumpGraph((GraphWidget) client_data,
	       EPSdumpLink,
	       EPSdumpNode,
	       EPSProlog,
	       EPSEpilog,
	       EPSComment,
	       (void *)file);

  (void)printf("Done.\n");
  if ((fclose(file)) != 0) {
    (void)fprintf(stderr, "fclose failed in DumpRef.\n");
  }
}


/*  */
/**********************************************************************
 * Function: Widget xdumpref(reference_structure *ref)
 *
 * Produces a reference structure dump in a window.
 *
 * Modifications:
 *      <list mods with name and date>
 */
Widget xdumpref(ref)
  reference_structure *ref;
{
  Widget panew, boxw, shellw, vieww, compw, quitw, dumpw, targetw, templatew;
  Dimension maxwidth = 0;
  struct GraphCallData *graphs = NULL;
  struct RefWidgets *refs = NULL;
  Dimension targetheight, targetwidth;
  Dimension templheight, templwidth;
  Dimension viewheight, viewwidth;
  Dimension templtitleheight, targettitleheight;
  Dimension contentsheight, contentswidth;
  XtArgVal targetgraph, templategraph;
  XtArgVal fontname;
  GNode target, template;

  /* initialisations so we don't get garbage */
  targetheight = targetwidth = templheight = templwidth = 0;
  viewheight = viewwidth = templtitleheight = targettitleheight = 0;
  contentsheight = contentswidth = 0;

  /* Check things first... */
  if (ref == (reference_structure *)0) {
    (void)fprintf(stderr, "Attempt to dump NULL reference structure.\n");
    return NULL;
  }
  /* Traverse and dump the target. */
  target.node.value = NULL;
  target.node.flags = 0;
  target.node.children = NULL;
  target.node.x = SPACEX / 2;
  target.node.y = -SPACEY;
  target.num = 0;
  target.used = 0;
  targetheight = SPACEY + stepidentries(ref->target, 20, 2, 
		 (indentrec *) 0, &target);
  targetwidth = (XtArgVal) GraphWidth(target.node.children[0]) + SPACEX;
  targetgraph = (XtArgVal) target.node.children[0];

  /* Traverse and dump the template. */
  template.node.value = NULL;
  template.node.flags = 0;
  template.node.children = NULL;
  template.node.x = SPACEX / 2;
  template.node.y = -SPACEY;
  template.num = 0;
  template.used = 0;
  templheight = SPACEY +
    stepidentries(ref->template, 20, 2, (indentrec *) 0, &template);
  templwidth = (XtArgVal) GraphWidth(template.node.children[0]) + SPACEX;
  templategraph = (XtArgVal) template.node.children[0];

  maxwidth = Max(targetwidth, templwidth);

  if (templwidth > MAXWIDTH || targetwidth > MAXWIDTH ||
      templheight > MAXHEIGHT || targetheight > MAXHEIGHT) {
    viewwidth = MAXWIDTH;
    viewheight = MAXHEIGHT;
  }
  /* Create the widgets */
  shellw = XtVaCreatePopupShell((String) makeshelltitle(ref),
				wmShellWidgetClass,
				toplevel,
				"title", (XtArgVal) makeshelltitle(ref),
				NULL);

  panew = XtVaCreateManagedWidget("Ref Paned",
				  panedWidgetClass,
				  shellw,
				  NULL);

  /* Do the command button section. */
  boxw = XtVaCreateManagedWidget("Command pane",
				 boxWidgetClass,
				 panew,
				 XtNshowGrip, FALSE,
				 NULL);

  quitw = XtVaCreateManagedWidget("CloseButton",
				  commandWidgetClass,
				  boxw,
				  XtNx, MARGIN,
				  XtNy, MARGIN,
				  XtNlabel, (XtArgVal) "Close Window",
				  XtNshapeStyle, (XtArgVal) XmuShapeOval,
				  NULL);

  dumpw = XtVaCreateManagedWidget("DumpButton",
				  commandWidgetClass,
				  boxw,
				  XtNx, 210,
				  XtNy, MARGIN,
				  XtNlabel, (XtArgVal) "Dump on File",
				  XtNshapeStyle, (XtArgVal) XmuShapeOval,
				  NULL);

  /* Do the graph section. */
  vieww = XtVaCreateManagedWidget("Scroller",
				  viewportWidgetClass,
				  panew,
				  XtNx, MARGIN,
				  XtNy, 0,
				  XtNwidth, viewwidth,
				  XtNheight, viewheight,
				  XtNallowVert, (Boolean) TRUE,
				  XtNallowHoriz, (Boolean) TRUE,
				  XtNlength, 500,
				  XtNborderWidth, (Dimension) 0,
				  NULL);

  /* assign layout widget */
  contentswidth = maxwidth + 2 * MARGIN;
  contentsheight = MARGIN + 40 + templheight + targetheight + 2 * MARGIN;
  compw = XtVaCreateManagedWidget("Contents",
				  compositeWidgetClass,
				  vieww,
				  XtNwidth, contentswidth,
				  XtNheight, contentsheight,
				  XtNresize, FALSE,
				  NULL);

  /* Make a title for the template graph. */
  (void)XtVaCreateManagedWidget("TemplateTitle",
				labelWidgetClass,
				compw,
				XtNx, MARGIN,
				XtNy, MARGIN,
				XtNwidth, maxwidth,
				XtNheight, templtitleheight,
				XtNlabel, (XtArgVal) "Template",
				XtNborderWidth, 0,
				NULL);

  /* assign graph widget */
  templatew = XtVaCreateManagedWidget("Contents",
				      graphWidgetClass,
				      compw,
				      XtNx, 
				        (Dimension)(maxwidth-templwidth)/
				        (Dimension)2,
				      XtNy, MARGIN + 20,
				      XtNwidth, templwidth,
				      XtNheight, templheight,
				      XtNgraph, templategraph,
				      XtNlayout, (Boolean) Done,
				      XtNdrawNode, (XtArgVal) DrawRefNode,
				      XtNnodeHit, (XtArgVal) ack,
				      XtNdrawLink, (XtArgVal) drawlink,
				      NULL);

  /* Make a title for the target graph. */
  (void)XtVaCreateManagedWidget("TargetTitle",
				labelWidgetClass,
				compw,
				XtNx, MARGIN,
				XtNy, MARGIN + 20 + templheight,
				XtNwidth, maxwidth,
				XtNheight, targettitleheight,
				XtNlabel, (XtArgVal) "Target",
				XtNborderWidth, 0,
				NULL);

  /* assign graph widget */
  targetw = XtVaCreateManagedWidget("Contents",
				    graphWidgetClass,
				    compw,
				    XtNx, 
				      (Dimension)(maxwidth-targetwidth)/
				      (Dimension)2,
				    XtNy, MARGIN + 40 + templheight,
				    XtNwidth, targetwidth,
				    XtNheight, targetheight,
				    XtNgraph, targetgraph,
				    XtNlayout, (Boolean) Done,
				    XtNdrawNode, (XtArgVal) DrawRefNode,
				    XtNdrawLink, (XtArgVal) drawlink,
				    NULL);

  /* set a couple of fonts */
  fontname = CVfont("5x7");
  if (fontname != (XtArgVal)0) {
    XtVaSetValues(quitw, XtNfont, fontname, NULL);
    XtVaSetValues(dumpw, XtNfont, fontname, NULL);
  }
  graphs = (struct GraphCallData *)malloc(sizeof(struct GraphCallData));
  if (graphs == (struct GraphCallData *)0) {
    fatal_error("%s\n", "malloc failed in xdumpref");
  }
  graphs->target = (GNode *) target.node.children[0];
  graphs->template = (GNode *) template.node.children[0];

  XtAddCallback(shellw, XtNdestroyCallback, whack_widl, (XtPointer)ref);
  XtAddCallback(shellw, XtNdestroyCallback, closeref, (XtPointer) graphs);
  XtAddCallback(quitw, XtNcallback, destroywid, (XtPointer)shellw);

  refs = (struct RefWidgets *)malloc(sizeof(struct RefWidgets));
  if (refs == (struct RefWidgets *)0) {
    fatal_error("%s\n", "malloc failed in xdumpref");
  }
  refs->target = (GraphWidget) targetw;
  refs->template = (GraphWidget) templatew;
  XtAddCallback(dumpw, XtNcallback, DumpRef, (XtPointer) refs);
  XtAddCallback(dumpw, XtNdestroyCallback, whack_refs, (XtPointer)refs);

  XtPopup((Widget) shellw, XtGrabNone);

  free((char *)target.node.children);
  free((char *)template.node.children);

  return shellw;
}


static GNode *BuildHistoryGraph(obj,w)
     label *obj;
     Widget w;
{
  GNode *graph;
  int x = 7;
  int y = 6;
  int flag = 0;

  if (w)
    flag++;

  graph = CreateNode((GNode *) 0, (void *)obj->vs);
  if (!BuildHistory(graph, obj, flag) || !graph->node.children) {
    popup_message(POPUP_MESS_INFO, "%s %d", 
		  "Couldn't build history graph for ", obj->vs);
    /* Free graph ? */
    return NULL;
  }

  if (w)
    XtVaGetValues(w, XtNstartX, &x, XtNstartY, &y, NULL);
      
  graph->node.x = x;
  graph->node.y = y;

  /* Compute the width and height. */
  ComputeLayout((GraphNode *) graph, LeftRight, 16, 15);

#if 0
  /* Invert the graph so that it grows from right to left. */
  graph->node.x = GraphWidth((GraphNode *)graph) + 15 + 15;
  ComputeLayout((GraphNode *) graph, LeftRight, -16, 15);
#endif

  return graph;
}

/*  */
/*ARGSUSED*/
/**********************************************************************
 * Function: static void component_rollback(GraphWidget widget, GNode *node)
 *
 * Pick a new version from component history object.
 *
 * Modifications:
 * 	<list mods with name and date>
 */
static void component_rollback(widget, node)
  GraphWidget widget;
  GNode *node;
{
  reference_structure *ref;
  label obj;
  label lbl;
  label bt;
  label root;
  GNode *graph;

  XtVaGetValues((Widget)widget,
                XtNgraph,&graph,
		XtNgraphValue,&ref,
		NULL);

  if (GLI_GETLINKITEM(&(ref->bt), BTBL_ROOT, 1, &lbl)) {
    short_error("The binding table has no root\n");
    return;
  }

  obj.vs = (int) graph->node.value;
  obj.inst = (graph == node) ? -1 : (int) node->node.value;

  switch (obj.inst) {
  case 0:
    (void)fprintf(stderr,"Retrieve hidden history.\n");
    obj.inst = -1;
    if ((graph = BuildHistoryGraph(&obj, widget))) {
      closehist((Widget)0,(XtPointer)widget,(XtPointer)0);
      XtVaSetValues((Widget) widget, XtNgraph, (XtArgVal)graph, NULL);
      ClearToBackground(widget);
      widget->graph.layout = NotDone;
      ClearGraphPart((GraphWidget )0, widget);
    }
    (*graphClassRec.core_class.expose)((Widget)widget, (XEvent *)0, (Region)0);
    return;
  case -1:
    (void)GC_GETCURRENT(&obj);
    (void)printf("Putting up the \"current\" composition version.\n");
    (void)UBL_UNBINDLABEL(copylbl(&root,&(ref->root)));
    *((int *)&(widget->graph.current)) = obj.vs;
    (void)buildref(&root, ref->template->idname, ref);
    return;
  default:
    (void)fprintf(stderr,"Retrieving version %d.\n", obj.inst);
    break;
  }

  if (!EDITED(&(ref->bt))) {
    popup_message(POPUP_MESS_INFO, "%s\n%s",
		  "A transient binding table version is",
		  "made for managing component rollback.");
    (void)copylbl(&bt,&(ref->bt));
    (void)CPO_COPYOBJ(&(ref->bt),&bt);
    newlastcmd(-(COPYOP+1),&(ref->bt),0,(char *)0,(char *)0,0,
	       (attrval *)0, &bt);
  }
  if (lbl.vs == obj.vs) {
    if (!UL_UNLINK(&(ref->bt), BTBL_ROOT, &lbl))
      (void)LINK(&(ref->bt), BTBL_ROOT, &obj, 1);
  } else {
    lbl.vs = obj.vs;
    lbl.inst = -1;
    (void)UL_UNLINK(&(ref->bt), BTBL_PARTS, &lbl);
    (void)LINK(&(ref->bt), BTBL_PARTS, &obj,1);
  }

  *((int *)&(widget->graph.current)) = obj.inst;
  (void)buildrefbt(ref->template->idname,&(ref->bt),TRUE,ref);
  (*graphClassRec.core_class.expose)((Widget)widget, (XEvent *)0, (Region)0);
}


/*  */
/*ARGSUSED*/
/**********************************************************************
 * Function: static void composition_rollback(GraphWidget widget, GNode *node)
 *
 * Pick a new version of the composite object.
 *
 * Modifications:
 * 	<list mods with name and date>
 */
static void composition_rollback(widget, node)
  GraphWidget widget;
  GNode *node;
{
  GNode *graph;
  reference_structure *oldref;
  label bt;
  label root;

  XtVaGetValues((Widget)widget,
		XtNgraph, &graph,
		XtNgraphValue, &oldref,
		NULL);

  bt.vs = (int)graph->node.value;
  bt.inst = (graph == node) ? -1 : (int) node->node.value;

  switch  (bt.inst) {
  case 0:
    (void)fprintf(stderr,"Retrieving hidden history.\n");
    bt.inst = -1;
    if ((graph = BuildHistoryGraph(&bt, widget))) {
      closehist((Widget)0,(XtPointer)widget,(XtPointer)0);
      XtVaSetValues((Widget) widget, XtNgraph, (XtArgVal)graph, NULL);
      ClearToBackground(widget);
      widget->graph.layout = NotDone;
      ClearGraphPart((GraphWidget )0, widget);
    }
    break;
  case -1:
    (void)printf("Retrieving the \"current\" version.\n");
    (void)UBL_UNBINDLABEL(copylbl(&root,&(oldref->root)));
    *((int *)&(widget->graph.current)) = oldref->bt.vs;
    (void)buildref(&root, oldref->template->idname, oldref);
    break;
  default:
    (void)printf("Retrieving version %d.\n", bt.inst);
    *((int *)&(widget->graph.current)) = bt.inst;
    (void)buildrefbt(oldref->template->idname, &bt, TRUE, oldref);
    break;
  }
  (*graphClassRec.core_class.expose)((Widget)widget, (XEvent *)0, (Region)0);
}

/*  */
/**********************************************************************
 * Function: Widget dump_temporal_history(label *obj, reference_structure *ref)
 *
 * Modifications:
 *      <list mods with name and date>
 */
Widget dump_temporal_history(obj, ref, comp)
  label *obj;
  reference_structure *ref;
  int comp;	/* 0 = composition, 1 = component. */
{
  Widget panew, historyw, vieww, boxw, shellw, quitw, dumpw;
  GNode *graph;
  Dimension width = 0;
  Dimension height = 0;
  XtArgVal fontname;
  char *p;
  void (*nodehit)();

  if (!(graph = BuildHistoryGraph(obj,NULL)))
    return NULL;

  if (GraphWidth((GraphNode *) graph) * 2 > MAXWIDTH)
    width = MAXWIDTH;
  if (GraphHeight((GraphNode *) graph) * 2 > MAXHEIGHT)
    height = MAXHEIGHT;

  if (!comp) {
    p = "Temporal History";
    nodehit = composition_rollback;
  } else {
    p = "Component Temporal History";
    nodehit = component_rollback;
  }

  shellw = XtVaCreatePopupShell((String)p,
				(WidgetClass) wmShellWidgetClass,
				(Widget) toplevel,
				"title", (String)p,
				NULL);

  panew = XtVaCreateManagedWidget("Pane",
				  panedWidgetClass,
				  shellw,
				  NULL);

  boxw = XtVaCreateManagedWidget("CloseButton",
				 boxWidgetClass,
				 panew,
				 XtNshowGrip, FALSE,
				 NULL);

  quitw = XtVaCreateManagedWidget("CloseButton",
				  commandWidgetClass,
				  boxw,
				  XtNx, MARGIN,
				  XtNy, MARGIN,
				  XtNwidth, 80,
				  XtNheight, 0,
				  XtNlabel, (XtArgVal) "Close Window",
				  XtNshapeStyle, (XtArgVal) XmuShapeOval,
				  NULL);

  dumpw = XtVaCreateManagedWidget("DumpButton",
				  commandWidgetClass,
				  boxw,
				  XtNx, 210,
				  XtNy, MARGIN,
				  XtNwidth, 80,
				  XtNheight, 0,
				  XtNlabel, (XtArgVal) "Dump on File",
				  XtNshapeStyle, (XtArgVal) XmuShapeOval,
				  NULL);

  vieww = XtVaCreateManagedWidget("Scroller",
				  viewportWidgetClass,
				  panew,
				  XtNx, 0,
				  XtNy, 0,
				  XtNwidth, width,
				  XtNheight, height,
				  XtNallowVert, (Boolean) TRUE,
				  XtNallowHoriz, (Boolean) TRUE,
				  XtNlength, 500,
				  XtNborderWidth, (Dimension) 0,
				  NULL);

  historyw = XtVaCreateManagedWidget("History",
				     graphWidgetClass,
				     vieww,
				     XtNx, MARGIN, 
				     XtNy, 0,
				     XtNwidth, 0,
				     XtNheight, 0,
				     XtNgraph,(XtArgVal)graph,
				     XtNscale, FIX_OFFSET * 2,
				     XtNspaceX, 16,
				     XtNspaceY, 15,
				     XtNnodeHit, (XtArgVal) nodehit,
				     XtNgraphValue, (XtArgVal) ref,
				     XtNnodeSize, 13, 
				     NULL);

  if (ref && ref->flags.bound) {
    *((int *)&(((GraphWidget)historyw)->graph.current)) = ref->bt.inst;
  } else {
    *((int *)&(((GraphWidget)historyw)->graph.current)) = obj->vs;
  }

  fontname = CVfont("5x7");
  if (fontname != (XtArgVal)0) {
    XtVaSetValues(quitw, XtNfont, (XtArgVal) fontname, NULL);
    XtVaSetValues(dumpw, XtNfont, (XtArgVal) fontname, NULL);
  }

  XtAddCallback(shellw, XtNdestroyCallback, whack_widl, (XtPointer)ref);
  XtAddCallback(shellw, XtNdestroyCallback, closehist, (XtPointer)historyw);
  XtAddCallback(quitw, XtNcallback, destroywid, (XtPointer)shellw);
  XtAddCallback(dumpw, XtNcallback, DumpHistory, (XtPointer) historyw);

  XtPopup((Widget) shellw, XtGrabNone);
  return shellw;
}

/*  */
/*ARGSUSED*/
/**********************************************************************
 * Function: static void freeMyValue(GNode *node)
 *
 * free's up the string and structure of a MyValue structure.
 * passed in to the FreeTree function.
 *
 * Modifications:
 * 	<list mods with name and date>
 */
static void freeMyValue(node)
  GNode *node;
{
  struct MyValue *to_free = NULL;

  if (!node) {
    (void)fprintf(stderr, "freeMyValue: (warning) null input parameter.\n");
    return;
  }

  to_free = (struct MyValue *)node->node.value;
  free((char *)to_free->string);
  free((char *)to_free);
}

/*  */
/*ARGSUSED*/
/**********************************************************************
 * Function: static void ack(GraphWidget widget, GNode *node)
 *
 * Pick a new version of the composite object.
 *
 * Modifications:
 * 	<list mods with name and date>
 */
static void ack(widget, node)
  GraphWidget widget;
  GNode *node;
{
  XBell(XtDisplay((Widget)widget), 100);
}

/*  */
/* ARGSUSED */
/**********************************************************************
 * Function: static void drawlink(SEVEN PARAMETERS)
 * Parameters:
 *      GraphWidget graph;
 *      GC gc;
 *      int x1, y1, x2, y2, size;
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void drawlink(graph, gc, x1, y1, x2, y2, size)
  GraphWidget graph;
  GC gc;
  int x1, y1, x2, y2, size;
{
  XDrawLine(XtDisplay(graph), XtWindow(graph), gc, x1, y1, x2, y2);
}
