/* 
 * Linkoping Intelligent Communication of Knowledge System (LINCKS)
 *      Copyright (C) 1993, Lin Padgham, Ralph Rnnquist
 *       Department of Computer and Information Sciences
 *		University of Linkoping, Sweden
 *		    581 83 Linkoping, Sweden
 *		       lincks@ida.liu.se
 *
 * These collective LINCKS programs are free software; you can 
 * redistribute them and/or modify them under the terms of the GNU
 * General Public License as published by the Free Software Foundation,
 * version 2 of the License.
 *
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the programs; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *   MODULE:	autostart.c
 *
 *   SCCSINFO:	@(#)autostart.c	1.2 11/4/93
 *
 *   AUTHOR:	Ralph R|nnquist, long time ago.
 *
 *   DESCRIPTION:
 *	Low-level liblincks routine to establish a conenction between
 *	a workspace (application) and database server.
 *
 *   MODIFIED:
 *	< add modifications here with name and date >
 *
 * EXTERNALLY-CALLABLE ROUTINES FOUND IN THIS MODULE:
 * int LL_FindMonitorHost(char *host,int *sock);
 */

/**********************************************************************
 * INCLUDES
 **********************************************************************/
#ifdef SVR42
#include <sys/byteorder.h>
#endif /* SVR 4.2 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/tcp.h>
#ifdef SVR42
#include <netinet/in.h>
#endif /* SVR 4.2 */
#include <netdb.h>
#include <rpc/rpc.h>
#include "lincks.h"
#include "xconfig.h"
#include "systypes.h"

/**********************************************************************
 * EXTERNALLY-AVAILABLE DATA FOUND IN THIS MODULE:
 **********************************************************************/
/* none */

/**********************************************************************
 * EXTERNAL FUNCTIONS USED BY THIS MODULE:
 **********************************************************************/
/* linckshandler.c */
extern int InitSignals(/* int s */);

/* logws.c */
extern void LogMess(/* varargs */);

/* libc */
#ifdef sun
extern struct hostent *gethostbyname();
extern unsigned long inet_addr( /* char *cp */ );
#endif

/**********************************************************************
 * EXTERNAL DATA STRUCTURES USED BY THIS MODULE:
 **********************************************************************/
extern int LL_uid;
extern char *sys_errlist[];
extern int sys_nerr;
extern int errno;

/**********************************************************************
 * INTERNAL FUNCTIONS USED BY THIS MODULE:
 **********************************************************************/
/* none */

/**********************************************************************
 * LOCAL DEFINES, STRUCTS, TYPEDEFS, ETC.:
 **********************************************************************/
#define LOG_ERRSTRING	sys_errlist[(errno >= sys_nerr) ? 0 : errno]
#define LOG_MESS(b)	LogMess(LL_uid,"%s, %s",b,LOG_ERRSTRING)
#define LOGIN_ERR \
"Couldn't connect to server. Server not running or server host not reachable."

/**********************************************************************
 * INTERNAL (STATIC) DATA:
 **********************************************************************/
/* none */

/*  */
/************************************************************************
 * Function: int LL_FindMonitorHost(char *host, int *sock)
 *
 * Establishes socket connection to a monitor for the present database.
 *
 * Modifications:
 *	<list mods here with name and date>
 *
 */
int LL_FindMonitorHost(host, sock)
  char *host;
  int *sock;
{ 
  static struct sockaddr_in server;
  static struct sockaddr *s = (struct sockaddr *) &server;
  static struct linger lingertime = { 1, 60 }; /* 60 second linger time */
  static int opt = SO_DEBUG;
  static struct hostent *hp = NULL;
  static u_long addr = 0;

  enum clnt_stat st;
  int gotit;
  int i;

  /* Use HOST from run-time configuration.  */
  (void)strcpy(host, HOST);

  /* try as a internet dot address first */
  if ((addr = inet_addr(host)) == -1) {    /* network order */
    if ((hp = gethostbyname(host)) == (struct hostent *)0) {
      LOG_MESS(LOGIN_ERR);
      return FAIL;
    } else {
      addr = *((u_long *) hp->h_addr_list[0]);  /* network order */
    }
  }

  /* Run-time configuration. */
  server.sin_port = htons(((int) TCPIPNO));
  server.sin_family = AF_INET;		/* Always */
  server.sin_addr.s_addr = addr;

  /* Tickle the monitor process to check its existence.  */
  if (hp) 
    if ((st = callrpc(host,((u_long)MONITORPROGNO),((u_long)TCPIPNO),(u_long)0,
		      xdr_void,(char *)0,
		      xdr_void,(char *)0))) {
      (void)fprintf(stderr,"\nLooked for %s %ld %ld %ld\n",
		    host,((u_long)MONITORPROGNO),
		    ((u_long)TCPIPNO),(u_long)0);
      clnt_perrno(st);
      (void)fprintf(stderr,"\nNo monitor for %s on %s!\n",DBDIR,HOST);
      return FAIL;
    }

  /* Create a socket for communication and export socket handle.  */
  if ((*sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    LOG_MESS("workspace: Creating socket failed");
    return FAIL;
  }

  /*
   * Tie socket to server address. (Actually the netserver process).
   */
 for(gotit=0, i=3; i>0; i--)
   if (connect(*sock, s, sizeof(server)) != 0) 
     LOG_MESS("workspace: Connect socket to DBS failed");
   else {
     gotit = 1;
     break;
   }

  if ( gotit == 0 ) {
    (void)close(*sock);
    return(FAIL);
  }

  /* 
   * just in the case the server closes the connection really fast,
   * enable linger with a 60 second options on timeout.
   */
  if (setsockopt(*sock, SOL_SOCKET, SO_LINGER, 
		 (char *)&lingertime, sizeof(lingertime))< 0) {
    LOG_MESS("workspace: Set SO_LINGER failed");
    return FAIL;
  }
#ifdef IP_OPTIONS  /* until linux fixes their IP code */
  /*
   * Set socket options to avoid transfer delays. This forces every
   * transfer packet to be available as soon as it comes rather than
   * having the network to wait 100ms for additional bytes.
   */
  if (setsockopt(*sock,IPPROTO_TCP,TCP_NODELAY,(char *)&opt,sizeof(int))< 0) {
    LOG_MESS("workspace: Set TCP_NODELAY failed.\n");
    return FAIL;
  }
#else
  LOG_MESS("workspace: Warning - TCP_NODELAY not supported.\n");
#endif /* n IP_OPTIONS */


  if (InitSignals(*sock) == FAIL) {
    LOG_MESS("InitSignals failed");
  }
  return SUCCESS;
}
