/* 
 * Linkoping Intelligent Communication of Knowledge System (LINCKS)
 *      Copyright (C) 1993, Lin Padgham, Ralph Rnnquist
 *       Department of Computer and Information Sciences
 *		University of Linkoping, Sweden
 *		    581 83 Linkoping, Sweden
 *		       lincks@ida.liu.se
 *
 * These collective LINCKS programs are free software; you can 
 * redistribute them and/or modify them under the terms of the GNU
 * General Public License as published by the Free Software Foundation,
 * version 2 of the License.
 *
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the programs; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * MODULE NAME: 	linckshandler.c
 *
 * SCCSINFO:		@(#)linckshandler.c	1.4 12/15/93
 *
 * ORIGINAL AUTHOR(S):  Hans Gustavsson, PEN pum group. 1992-05-26
 *
 * MODIFICATIONS:
 *      1993-08-12 Martin Sjlin, Rewrote retrieval code from dbs to
 *		get a bunch of pen and thus change user level interface
 *		into using a mapfn. (see usfuns.c)
 *	1993-08-26 Martin Sjlin. To enable better handling of the
 *		extra oob byte(s) (two) needed when we have a buggy
 *		select we need to read two bytes from the stream.
 *      1993-11-24 Martin Sjlin. SCO does not seem to support
 *              FIONREAD ioctl on pipes. Added code to use non blocking
 *              i/o on pipe. Using POSIX semantics, not sure if SCO have
 *              it or not, might add in O_NDELAY, too.

 * DESCRIPTION:
 *  This module handels the signal SIGURG and the read of
 *  ``out-of-band'' data from the socket.
 *
 *********************************************************************
 * EXTERNALLY-CALLABLE ROUTINES FOUND IN THIS MODULE:
 *********************************************************************
 * int InitSignals(int s)
 * void register_pen_function()
 * void POLL_oob()
 */
/*********************************************************************
 * INCLUDES:
 *********************************************************************/
#include <stdio.h>
#include <sys/types.h>
#ifdef HAVE_F_SETOWN
#include <unistd.h>
#include <fcntl.h>
#endif
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#ifndef sequent
#include <malloc.h>
#endif	/* sequent */
#if !defined(SIOCATMARK) && defined(SVR4)
#include <sys/sockio.h>
#endif /* n SIOCATMARK && SVR4 */
#if !defined(FIONREAD) && defined(SVR4)
#include <sys/filio.h>
#endif /* n FIONREAD && SVR4 */
#if !defined(SIGURG) && defined(SIGPOLL) && defined(SCO)
#include <stropts.h>                      /* to get options for streamio */
#endif
#include "dbcodes.h"
#include "sunlincks.h"
#include "systypes.h"
#include "config.h"

/*********************************************************************
 * EXTERNALLY-AVAILABLE	DATA FOUND IN THIS MODULE:
 *********************************************************************/
/* none */

/*********************************************************************
 * EXTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
#ifdef sequent
extern void free();
#endif  /* sequent */
extern errcode PEN_RETRIEVE_MESSAGE( /* int (*mapfm)(), char *extra */ );

/*********************************************************************
 * EXTERNAL DATA STRUCTURES USED BY THIS MODULE:
 *********************************************************************/
extern int LL_socketno;

/*********************************************************************
 * LOCAL DEFINES, STRUCTS, TYPEDEFS, ETC.:
 *********************************************************************/
/* none */

/*********************************************************************
 * INTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
static RETSIGTYPE Handler();
static void default_PEN_oob( /* char oob */ );
static int  PEN_mapfn( /* char *extra, int index, int type, 
			      char *msg, int count, int len */ );

/*********************************************************************
 * INTERNAL (STATIC) DATA:
 *********************************************************************/
/* the function to call if we get a parallel editing notification */
static void (*pen_notify_function) () = default_PEN_oob;
static int PEN_pipe[2];
static int PEN_inited = 0;
#ifdef OOBSELECTBUG
static int oobxbyte = 0;
#endif
#ifdef PENDEBUG
static char *pentype[] = {
  "Unknown", "Pen Resolve", "Pen Check", "Pen Poll" 
};
#endif
/*  */
/**********************************************************************
 * Function: int InitSignals(int s)
 *
 * Sets up liblincks signals SIGURG
 *
 * Modifications:
 *      Sat Oct  9 19:28:05 1993 Martin Sjlin. Added check for SIGURG
 *                   and if not defined try SIGPOLL instead (SCO system).
 *      Thu Nov 11 10:20:05 1993 Martin Sjolin -  SCO uses SIGUSR1 instead
 *                   of SIGURG for OOB and SIGPOLL for async I/O.
 *      <list mods with name and date>
 */
int InitSignals(s)
  int s;
{
#ifndef HAVE_F_SETOWN
  int pid = getpid();
#endif
 /* Register signals to take care of */
#ifdef SIGURG
  if (Signal(SIGURG, Handler) == BADSIG) {
    perror("InitSignals: Cannot register signal SIGURG");
    return (FAIL);
  }
#else
#ifdef SCO
  if (Signal(SIGUSR1, Handler) == BADSIG) {
    perror("InitSignals: Cannot register signal SIGUSR1");
    return (FAIL);
  }
#else 
  if (Signal(SIGPOLL, Handler) == BADSIG) {
    perror("InitSignals: Cannot register signal SIGPOLL");
    return (FAIL);
  }
#endif /* n SCO */
#endif
#ifdef HAVE_F_SETOWN
  if (fcntl(s, F_SETOWN, getpid()) < 0) {
#elif defined(SIOCSPGRP) 
  if (ioctl(s, SIOCSPGRP, &pid) < 0) {  /* socket(7) */
#else /* SIOCSPGRP */  
  /* Experimental!. No confirmation that this works. msj */
  /*
   * If we do not have a complete emulation of sockets on top
   * of streams, and since the SIOCSPGRP might be missing too, we 
   * check if we have
   *  - S_RDBAND (any band > 0, check streamio(7) under Solaris ) 
   *    to set on socket, else
   *  - assume that urgent data is implemented as hi priority message
   *    and enable signal.
   * It is very possible that I am wrong for the first case, and that
   * we should enable both S_RDBAND and S_HIPRI, but we will find out
   * later. [msj]
   */
#if defined(S_RDBAND)
  if (ioctl(s, I_SETSIG, S_RDBAND) < 0) { /* streamio(7) */
#else 
  if (ioctl(s, I_SETSIG, S_HIPRI) < 0)  { /* streamio(7) */
#endif /* n S_RDBAND */
#endif /* n HAVE_F_SETOWN && n SIOCSPGRP */
    perror("InitSignals: Cannot set pid/ioctl for OOB.");
    return (FAIL);
  }
  return (SUCCESS);
}

/*  */
/* ARGSUSED */
/**********************************************************************
 * Function: static RETSIGTYPE Handler(int sig, int code, struct sigcontext *scp)
 *
 * Takes care of system signals and performs appropriate
 * actions on them.
 *
 * Uses a function registered with register_pen_function()
 *
 * Modifications:
 *      <list mods with name and date>
 */
#ifdef HAVE_BSD_SIGHANDLER
static RETSIGTYPE Handler(sig, code, scp)
  int sig, code;
  struct sigcontext *scp;
#else
static RETSIGTYPE Handler(sig)
  int sig;
#endif
{
#ifdef OOBSELECTBUG
  int atoob = 0;
#endif
  int savederrno = errno;
  char oob;

  switch (sig) {
  case SIGALRM:
    break;

#ifdef SIGURG
  case SIGURG:
#else
#ifdef SCO
  case SIGUSR1:
#else
  case SIGPOLL:
#endif /* n SCO */
#endif /* n SIGURG */
#ifndef HAVE_RELIABLE_SIGNALS
    /* Reinstall the handler */
    if (Signal(sig, Handler) == BADSIG) 
      perror("Handler: Cannot register signal SIGURG/SIGUSR1/SIGPOLL");
#endif /* n HAVE_RELIABLE_SIGNALS */
#ifdef PENDEBUG
    (void)printf("Handler socket number = %d\n",LL_socketno);
#endif
#ifdef OOBSELECTBUG
    if (ioctl( LL_socketno, SIOCATMARK, &atoob) < 0)
      perror("Handler: ioctl for SIOCATMARK");
#endif
    if (recv(LL_socketno, &oob, 1, MSG_OOB) != 1) {
      if (errno != EINVAL) {
	perror("Handler");
	break;
      }
      /*
       * The error value EINVAL is not documented for recv in SunOS
       * manuals. We assume it is a case of OOB signal overrun and
       * act accordingly.
       */
#ifdef OOBSELECTBUG
      /* According to my doc(s), EINVAL signals that no OOB data
       * are currently available in the stream, but possibly later
       * in the stream - ioctl SIOCATMARK tells that this is the
       * case. atoob != 0 then the stream has been read up to
       * place where the oob data starts and thus it is safe
       * to read the following scrap byte.
       */
      if ( atoob == 0 )
	perror("Handler: OOB Sent, but not yet in buffer.");
#endif
#ifdef PENDEBUG
      perror("Handler: overrun");
#endif
      oob = OOB_MESSAGE_CODE;
    } 
#if defined(OOBSELECTBUG) 
    else {
      long  size = 0;
      byte  buff[2];  

    /* this is to get around a bug in SunOS/BSD and Solaris 2.x handling of
     * OOB data. After reading an OOB byte, you need to read at least one
     * byte from the "in-band-data" (a scrap byte) which we send after
     * the oob data. This MIGHT mess up the normal stream, but since we
     * are using TCP_NODELAY on the socket it looks like it is working 
     * and the response when logging out is quick (again).
     *
     * if we received an oob earlier, then we can read these extra 
     * bytes (SPECBYTE and RCOOB) and  clear the exception status 
     * in the select in LL_ReceiveStatus.
     *
     * check if data is in stream - if not we can assume that we will not
     * get the extra bytes after the oob ... also, and yet more important
     * we do want to 'hang' in any recv here
     */
#ifdef PENDEBUG
      fprintf( stderr, "oobxbyte = %d, atoob = %d\n", oobxbyte, atoob);
#endif
      if ( atoob == 1 ) {
	if (ioctl(LL_socketno, FIONREAD, &size) < 0) 
	  perror("Handler: At mark: ioctl when checking bytes in buffer.");
	else if ( size < sizeof(buff) ) {
#ifdef PENDEBUG
	  (void)fprintf(stderr, 
			"Handler: not enough data (%d) to read scrap.\n", 
			size);
#endif
	  oobxbyte++;			/* count number of missed occasions */

        /* we got the extra bytes and we "are at mark" */
	} 
	else if (recv(LL_socketno,&buff[0],sizeof(buff),0) == sizeof(buff)) {
	  if (buff[0] != SPECBYTE || buff[1] != RCOOB) 
	    (void) fprintf( stderr,
			   "Handler: Have a messed up datastream with %d %d\n",
			   (int)buff[0], (int)buff[1]);
	}
	else 
	  perror("Handler: Failed to read extra bytes after OOB");
      } /* atoob == 1 */
#ifdef PENDEBUG
      else
	(void)fprintf( stderr, 
		      "Handler: not at OOB(mark, handled in ReceiveStatus.\n");
#endif /* PENDEBUG */
    } /* oobselectbug */
#endif /* OOBSELECTBUG && n OOBSELECTBUGX */

    if (oob != OOB_MESSAGE_CODE) {
      perror("Handler: Unknown oob code");
      break;
    }
    (*pen_notify_function)(oob);
    break;

  default:
    perror("Handler: Cannot handle signal");
  }

 errno = savederrno;
}

/*  */
/**********************************************************************
 * Function: void register_pen_function()
 *
 * This function has as an argument the function that should
 * be called by the library if parallel editing notification is
 * required. It takes a char argument.
 *
 * Modifications:
 *      <list mods with name and date>
 */
void register_pen_function(func_to_call)
  void (*func_to_call) (/* char oob */);
{
  if (!func_to_call)
    return;
  pen_notify_function = func_to_call;
}


/*  */
/* ARGSUSED */
/**********************************************************************
 * Function: static void default_PEN_oob()
 * 
 * put an event in the event_queue
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void default_PEN_oob(oob)
  char oob;
{
  if (!PEN_inited) {
    if (pipe(PEN_pipe) != SUCCESS)
      perror("default_PEN_oob: Can't open pipe\n");
    PEN_inited = 1;
#ifdef PENDEBUG
    (void)fprintf( stderr," Open PEN pipe\n");
#endif
 }
  if(write(PEN_pipe[1], &oob, 1) != 1) 
    perror("default_PEN_oob: Write on pipe failed.");
#ifdef PENDEBUG
  else
   (void)fprintf( stderr, "Wrote on PEN pipe\n");
#endif
}


/*  */
/**********************************************************************
 * Function: void POLL_oob()
 * 
 * Polls the PEN_pipe to manage OOB signals from dbs. This function is
 * called from almost every liblincks/usfuns.c function.
 *
 * Modifications:
 *      1993-08-12 Martin Sjlin -  rewrote user level function to use
 *               a mapfn function as we retrieve a whole bunch of pens
 *               at once. Implied changes here.
 *
 * If this routine is not setup ( when default_PEN_oob is called the 
 * first time), just exit. Else, we flush the pipe and count the 
 * number of characters (it does not really care as we read all messages
 * from the dbs), then retrieve all messages use PEN_RETRIEVE_MESSAGE
 * and our private mapping function.                
 */
void POLL_oob()
{
  long pens,size = 0;
  int  count;
  int  st;
#ifdef NO_FIONREAD_PIPE
  int  flags;
#endif /* NO_FIONREAD_PIPE */
# define BUFSIZE 512
  static char buff[BUFSIZE]; /* skip the contents */

  if (!PEN_inited)
    return;

#ifdef NO_FIONREAD_PIPE
  /* It seems like you cannot do a FIONREAD on a pipe under SCO !*/ 
  if ((flags = fcntl(PEN_pipe[0], F_GETFL, 0)) < 0) {
    perror("POLL_oob: fcntl(1) failed to get on pipe");
    return;
  }
# ifdef O_NONBLOCK
  if (fcntl(PEN_pipe[0], F_SETFL, flags | O_NONBLOCK) < 0) {
# else
  if (fcntl(PEN_pipe[0], F_SETFL, flags | O_NDELAY) < 0) {
# endif /* n O_NONBLOCK */
    perror("POLL_oob: fcntl(2) failed to set on pipe");
    return;
  }
  size = 64;				/* try some large number */
#else
  /* we get number character available in pipe and read/skip them */
    if ( ioctl(PEN_pipe[0],FIONREAD,&size) < 0) 
      perror("POLL_oob: ioctl failed");
    if ( size <= 0)
      return;
#endif /* !NO_FIONREAD_PIPE */

  /* OK, skip number of character in pipe */
  for(pens = 0; pens < size; pens += count) {
    count = size < BUFSIZE ? size : BUFSIZE;
#ifdef NO_FIONREAD_PIPE
    if ((count = read( PEN_pipe[0], buff, count)) > 0)
      continue;
    else
# ifdef O_NONBLOCK
    if (count == -1 && errno == EAGAIN) break;
# else /* to handle SysV(O_NDELAY) (0) and 4.2BSD(O_NDELAY) (-1) */
    if (count <=  0 && errno == EWOULDBLOCK) break;
# endif /* n O_NONBLOCK */
    else {
      perror("POLL_oob: non-blocking read on pipe failed");
      goto exit;
    } 
#else /* NO_FIONREAD_PIPE */
    if ( read( PEN_pipe[0], buff, count) != count ) {
      perror("POLL_oob: read failed");
      goto exit;
    }
#endif /* n NO_FIONREAD_PIPE */
  } /* for */

  if ( (st = PEN_RETRIEVE_MESSAGE(PEN_mapfn, (char *)&pens)) != SUCCESS ) 
    (void)fprintf(stderr,"POLL_obb: Can't retrieve/treat messages (%d)\n", st);

 exit:
#ifdef NO_FIONREAD_PIPE
  if (fcntl(PEN_pipe[0], F_SETFL, flags) < 0) 
    perror("POLL_oob: fcntl(3) failed to set on pipe");
#endif /* NO_FIONREAD_PIPE */
  return;
}

/*  */
#ifndef PENDEBUG
/* ARGSUSED */
#endif
/**********************************************************************
 * Function: static int PEN_mapfn(char *extra, int index, int type, char *msg, int count, int len)
 *
 * Is called once for every pen message - for parameter description
 * see usfuns.c. Is always called at least once, for setup. Returns
 * nonzero value and it will no longer be call during processing 
 *
 */
static int PEN_mapfn( extra, index, type, msg, count, len)
 char *extra;
 int   index;
 int   type;
 char *msg; 
 int   count; 
 int   len;
{
#ifdef PENDEBUG
  int pens = (int) *extra;

  /* setup time ... */
  if ( count == 0 ) {
      (void)fprintf( stderr, 
	      "PEN_mapfn: Number of OOB in pipe: %d from dbs %d\n", 
	      pens, len);
  } else {
    if (type < 0 && type > 2) 
      type = 0;
    else 
      type++;

    if ( msg != NULL && msg[0] != '\0' )
      (void)fprintf( stderr, 
	      "PEN_mapfn: %d(%d): index %d type %s %s\n",
	      count, len, index, pentype[type], msg);
    else
      (void)fprintf( stderr, 
	      "PEN_mapfn: %d(%d): index %d type %s\n",
	      count, len, index, pentype[type]);

    if ( msg )
      (void) free(msg); /* allocated in rdwrmol.c in ReceiveImage */
  }

  /* and we like to see all of them ... */
  return(0);
#else
  return(1); /* see none */
#endif
}
