/* 
 * Linkoping Intelligent Communication of Knowledge System (LINCKS)
 *      Copyright (C) 1993, Lin Padgham, Ralph Rnnquist
 *       Department of Computer and Information Sciences
 *		University of Linkoping, Sweden
 *		    581 83 Linkoping, Sweden
 *		       lincks@ida.liu.se
 *
 * These collective LINCKS programs are free software; you can 
 * redistribute them and/or modify them under the terms of the GNU
 * General Public License as published by the Free Software Foundation,
 * version 2 of the License.
 *
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the programs; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* 
 * MODULE NAME: 	logging.c
 *
 * SCCSINFO:		@(#)logging.c	1.1 9/27/93
 *
 * ORIGINAL AUTHOR(S):  ????, Feb, 1987
 *
 * MODIFICATIONS:
 *	<list mods with name and date>
 *
 * DESCRIPTION:
 *  This file contains logging stuff.  Used to be in logmsg.c and
 *  logfile.c.
 */
/*********************************************************************
 * INCLUDES:
 *********************************************************************/
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <varargs.h>
#include "lincks.h"
#include "xconfig.h"
#include "systypes.h"

/*********************************************************************
 * EXTERNALLY-CALLABLE ROUTINES FOUND IN THIS MODULE:
 *********************************************************************/
void LogMessage(/* va_alist */);
void LogFile(/* va_alist */);

/*********************************************************************
 * EXTERNALLY-AVAILABLE	DATA FOUND IN THIS MODULE:
 *********************************************************************/
/* none */

/*********************************************************************
 * EXTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
/* xconfig.c */
extern char *LL_GetConfig(/* char *tag */);
extern long LL_GetConfigl(/* char *tag */);

/* libc */
#ifdef sun
extern time_t time();
#endif

/*********************************************************************
 * EXTERNAL DATA STRUCTURES USED BY THIS MODULE:
 *********************************************************************/
/* none */

/*********************************************************************
 * LOCAL DEFINES, STRUCTS, TYPEDEFS, ETC.:
 *********************************************************************/
#define BIGBUFFERSIZE 8192

/*********************************************************************
 * INTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
/* none */

/*********************************************************************
 * INTERNAL (STATIC) DATA: 
 *********************************************************************/
/* none */

/*  */
/**********************************************************************
 * Function: void LogMessage(va_alist)
 *
 * Logs an error message on file
 * 
 * Modifications:
 *      <list mods with name and date>
 */
#ifdef lint
/*VARARGS2*//*ARGSUSED*/
void LogMessage(uid_, format_, va_alist)
    UID uid_;
    char *format_;
#else
void LogMessage(va_alist)
#endif
    va_dcl
{
    va_list ap;
    int uid;		/* Assuming sizeof(UID) <= sizeof(int) */
    char *format;

    long secs;
    struct tm *t, *localtime();
    char msg[256];
    char userid[8];

    /* Get arguments */
    va_start(ap);
    uid = va_arg(ap, int);
    format = va_arg(ap, char *);

    if (format == NULL)
	format = "(null)";

    /* Build message */
    (void)vsprintf(msg, format, ap);
    switch (uid) {
    case UNKNOWN:
	(void)strcpy(userid, "U");
	break;
    case SUPERUSER:
	(void)strcpy(userid, "S");
	break;
    default:
	(void)sprintf(userid, "%d", uid);
    }

    /* Get current time */
    (void)time(&secs);
    t = localtime(&secs);
	
    /* Print time followed by error message on file */
#ifdef DEBUG
    LogFile(NETSERVLOG, "%02d-%02d-%02d  %02d:%02d:%02d  Pid: %d User: (%s) %d  DB: %s\n                    %s\n\n",
		        t->tm_year, t->tm_mon+1, t->tm_mday,
			t->tm_hour, t->tm_min, t->tm_sec,
	   	        getpid(), userid, TCPIPNO, DBDIR, msg);

#else
    LogFile(NETSERVLOG, "%02d-%02d-%02d  %02d:%02d:%02d  User: (%s) %d  DB: %s\n                    %s\n\n",
		        t->tm_year, t->tm_mon+1, t->tm_mday,
			t->tm_hour, t->tm_min, t->tm_sec,
	   	        userid, TCPIPNO, DBDIR, msg);
#endif
    va_end(ap);
}
/*  */
/**********************************************************************
 * Function: void LogFile(va_alist)
 * 
 * Modifications:
 *      <list mods with name and date>
 */
#ifdef lint
/*VARARGS2*/
/*ARGSUSED*/
void LogFile(file_, format_, va_alist)
    char *file_, *format_;
#else
void LogFile(va_alist)
#endif    
    va_dcl
{
    va_list ap;
    char *file;
    char *format;
    char filename[MAXPATHLEN];

    FILE *fp;
    static char filebuf[BIGBUFFERSIZE];

    /* Get arguments */
    va_start(ap);
    file = va_arg(ap, char *);
    format = va_arg(ap, char *);

    /* Create full path-name for logfile. */
    (void)sprintf(filename, "%s/%s", LOGDIR, file);

    if ((fp = fopen(filename,"a")) != NULL) {
#ifdef _IOFBF
      (void)setvbuf( fp, filebuf, _IOFBF, BIGBUFFERSIZE);
#else
      (void)setbuffer( fp, filebuf, BIGBUFFERSIZE);
#endif
      (void)vfprintf(fp, format, ap);
    } else {
      (void)sprintf(filename, "%s/%s", LOGDIR, LINCKSLOG);
      if ((fp = fopen(filename,"a")) != NULL) {
	(void)fprintf(fp,"Failed to open log file for %s/%s\n", LOGDIR, file);
	(void)vfprintf(fp,format,ap);
      } else {
        (void)sprintf(filename, "/tmp/%s", LINCKSLOG);
        if ((fp = fopen(filename,"a")) != NULL) {
	  (void)fprintf(fp,"Failed to open log file for %s/%s\n",
			LOGDIR, file);
	  (void)fprintf(fp,"Failed to open log file for %s/%s\n",
			LOGDIR, LINCKSLOG);
	  (void)vfprintf(fp,format,ap);
	}
      }
    }
    (void)fclose(fp);

#ifdef DEBUG
    (void)vfprintf(stderr, format, ap);
#endif
    va_end(ap);
}
