/* 
 * Linkoping Intelligent Communication of Knowledge System (LINCKS)
 *      Copyright (C) 1993, Lin Padgham, Ralph Rnnquist
 *       Department of Computer and Information Sciences
 *		University of Linkoping, Sweden
 *		    581 83 Linkoping, Sweden
 *		       lincks@ida.liu.se
 *
 * These collective LINCKS programs are free software; you can 
 * redistribute them and/or modify them under the terms of the GNU
 * General Public License as published by the Free Software Foundation,
 * version 2 of the License.
 *
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the programs; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


/*
 * MODULE NAME: 	moveobject.c
 *
 * SCCSINFO:		@(#)moveobject.c	1.1 9/27/93
 *
 * ORIGINAL AUTHOR(S):  Johan Alm, ????
 *
 * MODIFICATIONS:
 *	<list mods with name and date>
 *
 * DESCRIPTION:
 *
 *   File containing the functions that copies a molecule from the
 *   source file (the one not compacted) the the destination file
 *   (the compacted one).Image,attributs and forward links are copied
 *   without any modifications.The backward lincks pointing at a
 *   not reached nod is removed.The molecule file entry is modified and
 *   copied to the new molecule file.
 *   All functions in this file contains the word "MOVE'.
 */
/*********************************************************************
 * INCLUDES:
 *********************************************************************/
#include <stdlib.h>
#ifndef sequent
#include <malloc.h>
#endif
#include "cutoff.h"
#include "dbserver.h"
#include "libshared.h"
#include "systypes.h"

/*********************************************************************
 * EXTERNALLY-CALLABLE ROUTINES FOUND IN THIS MODULE:
 *********************************************************************/
int MoveObject( /* LABEL label */ );

/*********************************************************************
 * EXTERNALLY-AVAILABLE	DATA FOUND IN THIS MODULE:
 *********************************************************************/

/*********************************************************************
 * EXTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
/* cutoffmain.c */
extern void fatal_error( /* int error, char *function, char *message */ );

/* files.c */
extern long LengthOfFile( /* int fd */ );
extern int datfileOfIndex( /* int which, int index, int archive */ );
extern int molfileOfIndex( /* int which, int index, int archive */ );

/* flag.c */
extern int is_flag_set( /* LABEL label, unsigned char flag */ );

/* pointers.c */
extern int GetEntry( /* int fd, SEEK pos, char *entry, unsigned size */ );
extern int GetFilNo( /* LABEL label */ );
extern int GetFilPos( /* LABEL label */ );
extern int ChangeFilPos( /* LABEL label, long pos */ );

/*********************************************************************
 * EXTERNAL DATA STRUCTURES USED BY THIS MODULE:
 *********************************************************************/
/* none */

/*********************************************************************
 * LOCAL DEFINES, STRUCTS, TYPEDEFS, ETC.:
 *********************************************************************/
/* none */

/*********************************************************************
 * INTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
static long move_datasection( /* ftblindex, srcpos, srcsize, which */ );

/*********************************************************************
 * INTERNAL (STATIC) DATA:
 *********************************************************************/

/*  */
/**********************************************************************
 * Function: int MoveObject(LABEL label)
 *
 * Modifications:
 *      <list mods with name and date>
 */
int MoveObject(label)
  LABEL label;			/* label to molecule to be moved */
{
  FILEMOL molentry;		/* tmp entry in molecule file */
  static long srcmolpos;	/* old molecule file position */
  static long dstmolpos;	/* new molecule file position */
  long dstdatpos;		/* new data section position */
  int filetableindex;
  int srcmolfd = -1;	/* molecule fd of file to be compacted */
  int dstmolfd = -1;	/* molecule fd of compacted file */

  if ((filetableindex = GetFilNo(label)) == FAIL)
    fatal_error(FATAL_GEN, "MoveObject", "couldn't get filetableindex");

  srcmolpos = GetFilPos(label);

  srcmolfd = molfileOfIndex(MOL, filetableindex, ARCHIVE);
  dstmolfd = molfileOfIndex(MOL, filetableindex, NEW);
  dstmolpos = LengthOfFile(dstmolfd);

  switch (srcmolpos) {
  case UNDEFINED_ENTRY:
    return UNDEFINED_ENTRY;
  case FAIL:
    fatal_error(FATAL_GEN, "MoveObject", "couldn't get srcmolpos");
  default:
    break;
  }

  if (GetEntry(srcmolfd, srcmolpos,(char *)&molentry,sizeof(FILEMOL)) == FAIL)
    fatal_error(FATAL_GEN, "MoveObject", "GetEntry failed");

  dstdatpos = move_datasection(filetableindex, molentry.flink_filepos,
			       molentry.flink_size, FORWARD);
  molentry.flink_filepos = dstdatpos;
  dstdatpos = move_datasection(filetableindex, molentry.blink_filepos,
			       molentry.blink_size, BACKWARD);
  molentry.blink_filepos = dstdatpos;
  dstdatpos = move_datasection(filetableindex, molentry.attr_filepos,
			       molentry.attr_size, ATTRIBUTES);
  molentry.attr_filepos = dstdatpos;
  dstdatpos = move_datasection(filetableindex, molentry.image_filepos,
			       molentry.image_size, IMAGE);
  molentry.image_filepos = dstdatpos;

  if ((write(dstmolfd, (char *)&molentry, sizeof(FILEMOL)))!= sizeof(FILEMOL))
    fatal_error(FATAL_IO, "MoveObject", (char *)0);

  if (ChangeFilPos(label, dstmolpos) == FAIL)
    fatal_error(FATAL_GEN, "MoveObject", "ChangeFilPos failed");
  return SUCCESS;
}


/*  */
/**********************************************************************
 * Function: static long move_datasection(FOUR PARAMETERS)
 * Parameters:
 *	int ftblindex;
 *	long srcpos;
 *	int srcsize;
 *	MOLENTRYFIELD which;
 *
 * Modifications:
 *      <list mods with name and date>
 */
static long move_datasection(ftblindex, srcpos, srcsize, which)
  int ftblindex;
  long srcpos;
  int srcsize;
  MOLENTRYFIELD which;
{
  PACKEDBUFF pbuff;
  long dstpos = -1;
  int srcfd, dstfd;

  srcfd = datfileOfIndex(DAT, ftblindex, ARCHIVE);
  dstfd = datfileOfIndex(DAT, ftblindex, NEW);

  if ((pbuff.buff = (char *)malloc((unsigned)(srcsize))) == NULL)
    fatal_error(FATAL_MALLOC, "move_datasection", "pbuff.buff");

  pbuff.size = srcsize;
  pbuff.curp = 0;

  if (GetEntry(srcfd, srcpos, pbuff.buff, (unsigned)pbuff.size) == FAIL)
    fatal_error(FATAL_IO, "move_datasection", (char *)0);

  if (which == BACKWARD) {
    LINKGROUP *group, *gp;
    LINKS *fp;
    LINKITEM *prev, *cur, *next;

    if ((UnpackGroup(&group, &pbuff)) == FAIL)
      fatal_error(FATAL_GEN, "move_datasection", "UnpackGroup failed");

    prev = (LINKITEM *) 0;
    for (gp = group; gp != (LINKGROUP *) 0; gp = gp->nextgroup) {
      for (fp = gp->links; fp != (LINKS *) 0; fp = fp->nextlinks) {
	prev = (LINKITEM *) 0;
	for (cur = fp->linkitem; cur != (LINKITEM *) 0; cur = next) {
	  next = cur->nextitem;
	  if ((cur->label != 0) && !is_flag_set(cur->label, CO_REACHED)) {
	    if (prev == (LINKITEM *) 0)
	      fp->linkitem = next;
	    else
	      prev->nextitem = next;
	    free((char *)cur);
	  } else {
	    prev = cur;
	  }
	}
      }
    }

    (void)PackGroup(group, &pbuff);	/* pack the group into buffer */
  }
  dstpos = LengthOfFile(dstfd);
  if (write(dstfd, pbuff.buff, pbuff.size) != pbuff.size)
    fatal_error(FATAL_IO, "move_datasection", (char *)0);

  free(pbuff.buff);
  return dstpos;
}
