/* 
 * Linkoping Intelligent Communication of Knowledge System (LINCKS)
 *      Copyright (C) 1993, Lin Padgham, Ralph Rnnquist
 *       Department of Computer and Information Sciences
 *		University of Linkoping, Sweden
 *		    581 83 Linkoping, Sweden
 *		       lincks@ida.liu.se
 *
 * These collective LINCKS programs are free software; you can 
 * redistribute them and/or modify them under the terms of the GNU
 * General Public License as published by the Free Software Foundation,
 * version 2 of the License.
 *
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the programs; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


/*
 * MODULE NAME: 	traverse.c
 *
 * SCCSINFO:		@(#)traverse.c	1.3 12/15/93
 *
 * ORIGINAL AUTHOR(S):  Tom Nilsson, ????
 *
 * MODIFICATIONS:
 *	<list mods with name and date>
 *
 * DESCRIPTION:
 *  File containing Traverse. Cutoff2 calls traverse.
 *  Traverse searches the nodestructure and sets the CO_REACHED flag to
 *  true for the nodes that can be reached from the systemnode and shall
 *  be moved to the new version of the database.
 */
/*********************************************************************
 * INCLUDES:
 *********************************************************************/
#ifndef sequent
#include <malloc.h>
#endif	/* sequent */
#ifdef SCO
#include <sys/types.h>
#endif
#include <sys/file.h>
#include <unistd.h>

#include "xconfig.h"
#include "cutoff.h"
#include "dbserver.h"
#include "monitor.h"
#include "libshared.h"
#include "systypes.h"

/*********************************************************************
 * EXTERNALLY-CALLABLE ROUTINES FOUND IN THIS MODULE:
 *********************************************************************/
int Traverse( /* */ );

/*********************************************************************
 * EXTERNALLY-AVAILABLE	DATA FOUND IN THIS MODULE:
 *********************************************************************/
/* none */

/*********************************************************************
 * EXTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
/* cutoffmain.c */
extern void fatal_error( /* int error, char *function, char *message */ );

/* files.c */
extern int molfileOfIndex( /* int index, int archive */ );
extern int datfileOfIndex( /* int index, int archive */ );

/* flag.c */
extern void set_flag( /* LABEL label, unsigned char flag */ );
extern int is_flag_set( /* LABEL label, unsigned char flag */ );

/* links.c */
extern void FreeLinks( /* LINKGROUP *lg */ );

/* pointers.c */
extern int GetEntry( /* int fd, SEEK pos, char *entry, unsigned size */ );
extern int GetDatPtr( /* int fd, SEEK pos, int field */ );
extern int GetDatSize( /* int fd, SEEK pos, int field */ );
extern int GetFilNo( /* LABEL label */ );
extern int GetFilPos( /* LABEL label */ );

#ifdef sequent
extern char *malloc();
extern void free();
#endif  /* sequent */

/*********************************************************************
 * EXTERNAL DATA STRUCTURES USED BY THIS MODULE:
 *********************************************************************/
/* cutoffmain.c */
extern long global_idxlen;

/*********************************************************************
 * LOCAL DEFINES, STRUCTS, TYPEDEFS, ETC.:
 *********************************************************************/
/* none */

/*********************************************************************
 * INTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
static int FetchNodesToCheck( /* LABEL node,LABEL *next */ );
static int SetReached( /* LABEL node, LABEL *next */ );
static int DoFetchNodesToCheck( /* FOUR PARAMETERS */ );

/*********************************************************************
 * INTERNAL (STATIC) DATA:
 *********************************************************************/
/* none */

/*  */
/**********************************************************************
 * Function: int Traverse()
 *
 * set CO_REACHED flag to true for all nodes that can be
 * reached from the systemnode and shall be moved.
 *
 * returns SUCCESS or FAIL
 *
 * FetchNodesToCheck,ConvToLabel
 *
 * Modifications:
 *      <list mods with name and date>
 */
int Traverse()
{
  LABEL label = FIRSTLABEL;
  LABEL nextlabel = 0;
  long max_diff = 0;
  long n_back_step = 0;
  long sum_diff = 0;
  int return_code;

  /* system node is reached */
  set_flag(FIRSTLABEL, CO_REACHED);
  while (label < ConvToLabel(global_idxlen)) {
    nextlabel = 0;
    if (is_flag_set(label, CO_REACHED) && !is_flag_set(label, CO_CHECK)) {
      set_flag(label, CO_CHECK);
      if ((return_code = FetchNodesToCheck(label, &nextlabel)) == FAIL) {
	(void)printf("Traverse: couldn't fetch nodes to check.\n");
	break;
      }
    }
    if (nextlabel != 0) {
      n_back_step++;
      if ((label - nextlabel) > max_diff)
	max_diff = label - nextlabel;
      sum_diff += (label - nextlabel);
      label = nextlabel;
    } else {
      label += sizeof(INDEX);
    }
  }
  if (return_code == FAIL)
    (void)printf("Last node processed: %d\n",label);
  (void)printf("Maximal back step length: %d\n",max_diff);
  (void)printf("Sum of back step lengths: %d\n",sum_diff);
  (void)printf("Number of back step times: %d\n",n_back_step);
  return return_code;
}


/*  */
/**********************************************************************
 * Function: static int FetchNodesToCheck(LABEL node,LABEL *next)
 *
 * A wrapper for DoFetchNodesToCheck to handle memory reclaim on all
 * returns.
 * Returns FAIL on fails and SUCCESS on success.
 *
 * Modifications:
 *      <list mods with name and date>
 */
static int FetchNodesToCheck(node, next)
  LABEL node;
  LABEL *next;
{
  PACKEDBUFF pbuff;
  LINKGROUP *group;
  long return_code;

  pbuff.buff = (char *)0;
  pbuff.size = 0;
  pbuff.curp = 0;
  group = (LINKGROUP *) 0;

  return_code = DoFetchNodesToCheck(node, next, &pbuff, &group);
  if (pbuff.buff != (char *)0)
    free(pbuff.buff);
  if (group != (LINKGROUP *) 0)
    FreeLinks(group);

  return return_code;
}

/*  */
/**********************************************************************
 * Function: static int SetReached(LABEL cur, LABEL node, LABEL *next)
 *
 * set CO_REACHED for node and keep least non-zero label in next.
 *
 * Modifications:
 *      <list mods with name and date>
 */
static int SetReached(cur, node, next)
  LABEL cur;
  LABEL node;
  LABEL *next;
{
  if ((node == 0) || is_flag_set(node, CO_REACHED))
    return SUCCESS;

  set_flag(node, CO_REACHED);
  if (node >= cur)
    return SUCCESS;

  if (((*next) == 0) || (node < (*next)))
    (*next) = node;
  return SUCCESS;
}

/*  */
/**********************************************************************
 * Function: static int DoFetchNodesToCheck(FOUR PARAMETERS)
 * Parameters:
 *	LABEL node
 *	LABEL *next
 *	PACKEDBUFF *pbuff
 *	LINKGROUP **group
 *
 * Follow the forward links from the node and mark all nodes referred to
 * as CO_REACHED. The least label of next and newly reached nodes is
 * assigned to next.
 * Return FAIL for fails, and SUCCESS on success.
 *
 * Modifications:
 *      <list mods with name and date>
 */
static int DoFetchNodesToCheck(node, next, pbuff, group)
  LABEL node;
  LABEL *next;
  PACKEDBUFF *pbuff;
  LINKGROUP **group;
{
  long filpos;
  int filno;
  int molfd;
  int datfd;
  long fowptr;
  int size;
  LINKGROUP *gp;
  LINKS *fp;
  LINKITEM *ip;

  if ((filpos = GetFilPos(node)) == UNDEFINED_ENTRY)
    return SUCCESS;

  if (filpos < 0)
    return FAIL;

  if ((filno = GetFilNo(node)) == FAIL)
    return FAIL;

  molfd = molfileOfIndex(MOL, filno, ARCHIVE);
  datfd = datfileOfIndex(DAT, filno, ARCHIVE);

  pbuff->buff = NULL;
  if (((fowptr = GetDatPtr(molfd, filpos, FORWARD)) == FAIL) ||
      ((size = GetDatSize(molfd, filpos, FORWARD)) < 0) ||
      ( size > 0 && ((pbuff->buff = malloc((unsigned)size)) == NULL)))
    return FAIL;
  pbuff->size = size;

  if (((GetEntry(datfd, fowptr, pbuff->buff, (unsigned)size)) == FAIL) ||
      (UnpackGroup(group, pbuff) == FAIL))
    return FAIL;

  for (gp = *group; gp != (LINKGROUP *) 0; gp = gp->nextgroup) {
    for (fp = gp->links; fp != (LINKS *) 0; fp = fp->nextlinks) {
      for (ip = fp->linkitem; ip != (LINKITEM *) 0; ip = ip->nextitem) {
	if ((SetReached(node, ip->label, next) == FAIL) ||
	    (SetReached(node, ip->vs, next) == FAIL))
	  return FAIL;
      }
    }
  }
  return SUCCESS;
}
