/* 
 * Linkoping Intelligent Communication of Knowledge System (LINCKS)
 *      Copyright (C) 1993, Lin Padgham, Ralph Rnnquist
 *       Department of Computer and Information Sciences
 *		University of Linkoping, Sweden
 *		    581 83 Linkoping, Sweden
 *		       lincks@ida.liu.se
 *
 * These collective LINCKS programs are free software; you can 
 * redistribute them and/or modify them under the terms of the GNU
 * General Public License as published by the Free Software Foundation,
 * version 2 of the License.
 *
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the programs; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * MODULE NAME: 	dbdump.c
 *
 * SCCSINFO:		@(#)dbdump.c	1.1 9/27/93
 *
 * ORIGINAL AUTHOR(S):  Ralph R|nnquist, 1991-02-07
 *
 * MODIFICATIONS:
 *	<list mods with name and date>
 *
 * DESCRIPTION:
 *	The dbdump program attempts to read a database and
 *	produce (on stdout) a t2lincks-loadable description
 *	of the current database state.
 *	NOTE: Bound labels are not treated accurately!!
 */
/*********************************************************************
 * INCLUDES:
 *********************************************************************/
#include <stdio.h>
#include <fcntl.h>
#ifndef sequent
#include <malloc.h>
#endif	/* sequent */
#include <string.h>

#include "systypes.h"
#include "liblincks.h"

/*********************************************************************
 * EXTERNALLY-CALLABLE ROUTINES FOUND IN THIS MODULE:
 *********************************************************************/
int main( /* int argc, char **argv */ );

/*********************************************************************
 * EXTERNALLY-AVAILABLE	DATA FOUND IN THIS MODULE:
 *********************************************************************/
/* none */

/*********************************************************************
 * EXTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
#ifdef __ultrix
extern char *strdup( /* char *incoming */ );
#endif

#ifdef sequent
extern char *malloc();
extern void free();
#endif  /* sequent */

/*********************************************************************
 * EXTERNAL DATA STRUCTURES USED BY THIS MODULE:
 *********************************************************************/
extern char *optarg;
extern int optind;
extern int opterr;
extern int global_comhistoff;

/*********************************************************************
 * LOCAL DEFINES, STRUCTS, TYPEDEFS, ETC.:
 *********************************************************************/
typedef struct {
  label lbl;
  char *group;
  char *field;
} mapaccess;

typedef struct x_entry {
  struct x_entry *next, *link;
  label lbl;
} DUMPENTRY;

#define MAXDEPTH	100
#define STRLEN		124
#define PRINTNODE(x)	if (x) (void)fprintf(outfile, "N%d",x); else (void) fprintf(outfile, "\"System Root\"")

/*********************************************************************
 * INTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
static void Printstr( /* char *p, int n */ );
static void Printval( /* attrval *v */ );
static int Attrfieldfn( /* mapaccess *mp, char *fld, int c, int n */ );
static int Attrgroupfn( /* mapaccess *mp, char *grp, int c, int n */ );
static void Dumpfirst( /* label *lbl */ );
static void Addtodumplist( /* label *lbl */ );
static int Itemmapfn( /* mapaccess *mp, label *item, int c, int n */ );
static int Fieldmapfn( /* mapaccess *mp, char *fld, int c, int n */ );
static int Groupmapfn( /* mapaccess *mp, char *grp, int c, int n */ );
static void Builddump();
static int Itemdumpfn( /* mapaccess *mp, label *item, int c, int n */ );
static int Fielddumpfn( /* mapaccess *mp, char *fld, int c, int n */ );
static int Groupdumpfn( /* mapaccess *mp, char *grp, int c, int n */ );
static void Dumplinks( /* DUMPENTRY *p */ );
static void arginit( /* int argc, char **argv */ );
static void usage();
static FILE *do_login();
static void print_version();

/*********************************************************************
 * INTERNAL (STATIC) DATA:
 *********************************************************************/
/* The savelist consists of all links to handle at a level.  */
static DUMPENTRY *savelist;

static int level = 0;
static int maxdepth = 0;

/*
 * Each dumplst entry is an object instance that 
 * should be dumped. The list is kept sorted by lbl.vs.
 */
static DUMPENTRY *dumplst = NULL;

static FILE *outfile = NULL;

static char dbdir[STRLEN];
static char username[STRLEN];
static char password[STRLEN];
static char filename[STRLEN];
static int noninteractive = 0;

/*  */
/**********************************************************************
 * Function: static void Printstr(char *p, int n)
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void Printstr(p, n)
  char *p;
  int n;
{
  (void)fprintf(outfile, "\"");
  while (n-- > 0) {
    if (*p != '\0') {
      if ((*p == '"') || (*p == '\n') || (*p == '\\'))
	(void)fprintf(outfile, "\\");
      (void)fprintf(outfile, "%c", *(p++));
    }
  }
  (void)fprintf(outfile, "\"");
}

/*  */
/**********************************************************************
 * Function: static void Printval(attrval *v)
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void Printval(v)
  attrval *v;
{
  (void)fprintf(outfile, " ");
  Printstr(v->attvalue, v->attsize);
  (void)fprintf(outfile, "\n");
}


/*  */
/* ARGSUSED */
/**********************************************************************
 * Function: static int Attrfieldfn(mapaccess *mp, char *fld, int c, int n)
 *
 * Modifications:
 *      <list mods with name and date>
 */
static int Attrfieldfn(mp, fld, c, n)
  mapaccess *mp;
  char *fld;
  int c;
  int n;
{
  attrval v;
  if (c == 0)
    return 0;

  if (GA_GETATTR(&(mp->lbl), mp->group, fld, &v) == SUCCESS) {
    (void)fprintf(outfile, "\t");
    Printstr(fld, strlen(fld));
    (void)fprintf(outfile, " :");
    Printval(&v);
    if (v.attsize > 0)
      free(v.attvalue);
  }
  return 0;
}

/*  */
/* ARGSUSED */
/**********************************************************************
 * Function: static int Attrgroupfn(mapaccess *mp, char *grp, int c, int n)
 *
 * Modifications:
 *      <list mods with name and date>
 */
static int Attrgroupfn(mp, grp, c, n)
  mapaccess *mp;
  char *grp;
  int c;
  int n;
{
  if (c == 0)
    return 0;

  (void)fprintf(outfile, ".group ");
  PRINTNODE(mp->lbl.vs);
  (void)fprintf(outfile, " ");
  Printstr(grp, strlen(grp));
  (void)fprintf(outfile, "\n");

  mp->group = strdup(grp);
  if ((GAN_GETATTRNAMES(&(mp->lbl), grp, Attrfieldfn, (char *)mp)) != SUCCESS)
    (void)fprintf(stderr, "Attrgroupfn: GAN_GETATTRNAMES failed\n");
  free(mp->group);
  mp->group = 0;

  (void)fprintf(outfile, "\n");
  return 0;
}

/*  */
/**********************************************************************
 * Function: static void Dumpfirst(label *lbl)
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void Dumpfirst(lbl)
  label *lbl;
{
  mapaccess mp;
  attrval v;

  (void)fprintf(stdout, "+ Node N%d\n", lbl->vs);
  if (lbl->vs == 0) {
    (void)fprintf(outfile,"########################################\n");
    (void)fprintf(outfile,"# object \"System Root\" is predefined.\n");
  } else {
    (void)fprintf(outfile,"########################################\n.object ");
    PRINTNODE(lbl->vs);
    v.attsize = 0;
    if ((GI_GETIMAGE(lbl, &v)) != SUCCESS)
      (void)fprintf(stderr, "Dumpfirst: GI_GETIMAGE failed\n");
    Printval(&v);
    if (v.attsize > 0)
      free(v.attvalue);
  }
  mp.lbl.vs = lbl->vs;
  mp.lbl.inst = lbl->inst;
  if ((GAGN_GETATTRGROUPNAMES(lbl, Attrgroupfn, (char *)&mp)) != SUCCESS)
    (void)fprintf(stderr, "Dumpfirst: GAGN_GETATTRGROUPNAMES failed\n");
}

/*  */
/**********************************************************************
 * Function: static void Addtodumplist(label *lbl)
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void Addtodumplist(lbl)
  label *lbl;
{
  DUMPENTRY **at, *p;

  at = &dumplst;

  while ((*at) && ((*at)->lbl.vs < lbl->vs))
    at = &((*at)->next);

  if ((*at) && ((*at)->lbl.vs == lbl->vs))
    return;

  p = (DUMPENTRY *) malloc(sizeof(DUMPENTRY));
  p->lbl.vs = lbl->vs;
  p->lbl.inst = lbl->inst;
  p->link = savelist;
  savelist = p;
  p->next = (*at);
  *at = p;

  Dumpfirst(lbl);
  return;
}

/*  */
/* ARGSUSED */
/**********************************************************************
 * Function: static int Itemmapfn(mapaccess *mp, label *item, int c, int n)
 *
 * Modifications:
 *      <list mods with name and date>
 */
static int Itemmapfn(mp, item, c, n)
  mapaccess *mp;
  label *item;
  int c;
  int n;
{
  if (c == 0)
    return 0;

  Addtodumplist(item);
  return 0;
}

/*  */
/* ARGSUSED */
/**********************************************************************
 * Function: static int Fieldmapfn(mapaccess *mp, char *fld, int c, int n)
 *
 * Modifications:
 *      <list mods with name and date>
 */
static int Fieldmapfn(mp, fld, c, n)
  mapaccess *mp;
  char *fld;
  int c;
  int n;
{
  if (c == 0)
    return 0;

  mp->field = strdup(fld);
  if ((GLV_GETLINKVAL(&(mp->lbl), mp->group,
		      fld, Itemmapfn, (char *)mp)) != SUCCESS)
    (void)fprintf(stderr, "Fieldmapfn: GLV_GETLINKVAL failed\n");
  free(mp->field);
  mp->field = 0;
  return 0;
}

/*  */
/* ARGSUSED */
/**********************************************************************
 * Function: static int Groupmapfn(mapaccess *mp, char *grp, int c, int n)
 *
 * Modifications:
 *      <list mods with name and date>
 */
static int Groupmapfn(mp, grp, c, n)
  mapaccess *mp;
  char *grp;
  int c;
  int n;
{
  if (c == 0)
    return 0;

  mp->group = strdup(grp);
  if ((GLN_GETLINKNAMES(&(mp->lbl), grp, Fieldmapfn, (char *)mp)) != SUCCESS)
    (void)fprintf(stderr, "Groupmapfn: GLN_GETLINKNAMES failed\n");
  free(mp->group);
  mp->group = 0;
  return 0;
}

/*  */
/**********************************************************************
 * Function: static void Builddump()
 *
 * Process the current savelist and make a new one for the next level.
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void Builddump()
{
  mapaccess mp;
  DUMPENTRY *oldsave;

  while (savelist) {
    oldsave = savelist;
    savelist = 0;
    level++;
    (void)fprintf(stderr, "Beginning level %d\n", level);
    (void)fprintf(outfile, "### Beginning level %d ###\n", level);
    while (oldsave) {
      if (level >= maxdepth) {
	(void)fprintf(stderr, "## Exceeding maximal depth for object ");
	(void)fprintf(stderr, "N%d ##\n", oldsave->lbl.vs);
	(void)fprintf(outfile, "#### Exceeding maximal depth for object ");
	PRINTNODE(oldsave->lbl.vs);
	(void)fprintf(outfile, "####.\n");
	oldsave->lbl.inst = oldsave->lbl.vs;
	oldsave->lbl.vs = -1;
      } else {
	mp.lbl.vs = oldsave->lbl.vs;
	mp.lbl.inst = oldsave->lbl.inst;
	mp.group = 0;
	mp.field = 0;
	if ((GLGN_GETLINKGROUPNAMES(&(oldsave->lbl), Groupmapfn,
				    (char *)&mp)) != SUCCESS)
	  (void)fprintf(stderr, "Builddump: GLGN_GETLINKGROUPNAMES failed\n");
      }
      oldsave = oldsave->link;
    }
  }
}

/*  */
/********************************************************************/
/* Dumping links */

/* ARGSUSED */
/**********************************************************************
 * Function: static int Itemdumpfn(mapaccess *mp, label *item, int c, int n)
 *
 * Modifications:
 *      <list mods with name and date>
 */
static int Itemdumpfn(mp, item, c, n)
  mapaccess *mp;
  label *item;
  int c;
  int n;
{
  if (c == 0)
    return 0;

  if (c == 1) {
    (void)fprintf(outfile, "\t");
    Printstr(mp->field, strlen(mp->field));
    (void)fprintf(outfile, " = ");
  } else {
    (void)fprintf(outfile, ", ");
  };
  PRINTNODE(item->vs);
  return 0;
}

/*  */
/* ARGSUSED */
/**********************************************************************
 * Function: static int Fielddumpfn(mapaccess *mp, char *fld, int c, int n)
 *
 * Modifications:
 *      <list mods with name and date>
 */
static int Fielddumpfn(mp, fld, c, n)
  mapaccess *mp;
  char *fld;
  int c;
  int n;
{
  if (c == 0)
    return 0;

  if (strcmp(fld, "Parent") == 0 && strcmp(mp->group, "SYSTEM") == 0)
    return 0;

  mp->field = strdup(fld);
  if ((GLV_GETLINKVAL(&(mp->lbl), mp->group, fld, 
		 Itemdumpfn, (char *)mp)) != SUCCESS)
    (void)fprintf(stderr, "Fielddumpfn: GLV_GETLINKVAL failed\n");
  free(mp->field);
  mp->field = 0;

  (void)fprintf(outfile, "\n");
  return 0;
}

/*  */
/**********************************************************************
 * Function: static int Groupdumpfn(mapaccess *mp, char *grp, int c, int n)
 *
 * Modifications:
 *      <list mods with name and date>
 */
static int Groupdumpfn(mp, grp, c, n)
  mapaccess *mp;
  char *grp;
  int c;
  int n;
{
  if (c == 0)
    return 0;

  (void)fprintf(outfile, ".links ");
  PRINTNODE(mp->lbl.vs);
  (void)fprintf(outfile, " ");
  Printstr(grp, strlen(grp));
  (void)fprintf(outfile, "\n");

  mp->group = strdup(grp);
  if ((GLN_GETLINKNAMES(&(mp->lbl), grp, Fielddumpfn, (char *)mp)) != SUCCESS)
    (void)fprintf(stderr, "Groupdumpfn: GLN_GETLINKNAMES failed\n");
  free(mp->group);
  mp->group = 0;

  if (c == n)
    (void)fprintf(outfile, "########################################\n\n");
  else
    (void)fprintf(outfile, "\n");

  return 0;
}

/*  */
/**********************************************************************
 * Function: static void Dumplinks(DUMPENTRY *p)
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void Dumplinks(p)
  DUMPENTRY *p;
{
  mapaccess mp;

  (void)fprintf(stderr, "*** Dumping the links ***\n");
  (void)fprintf(outfile, "######## Here come all links: #########\n\n");
  while (p != 0) {
    if ((mp.lbl.vs = p->lbl.vs) != -1) {
      mp.lbl.inst = p->lbl.inst;
      mp.group = 0;
      mp.field = 0;
      if ((GLGN_GETLINKGROUPNAMES(&(p->lbl), Groupdumpfn,
				  (char *)&mp)) != SUCCESS)
	(void)fprintf(stderr, "Dumplinks: GLGN_GETLINKGROUPNAMES failed\n");

    }
    p = p->next;
  }
}

/*  */
/**********************************************************************
 * Function: int main(int argc, char **argv)
 *
 * Modifications:
 *      <list mods with name and date>
 */
int main(argc, argv)
  int argc;
  char **argv;
{
  label root;

  global_comhistoff = 1;

  arginit(argc, argv);
  outfile = do_login();

  if (GSR_GETSYSTEMROOT(&root) != SUCCESS) {
    (void)fprintf(stderr, "Error: Couldn't get System Root !!!\n");
    exit(1);
  }
  Addtodumplist(&root);
  Builddump();
  Dumplinks(dumplst);

  (void)LOGOUT();
  return (0);
}

/*  */
/**********************************************************************
 * Function: static void arginit(int argc, char **argv)
 *
 * processes the command line options.
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void arginit(argc, argv)
  int argc;
  char **argv;
{
  int c = 0, user = 0, pword = 0;
  char *db;

  bzero(dbdir, sizeof(dbdir));
  bzero(username, sizeof(username));
  bzero(password, sizeof(password));
  bzero(filename, sizeof(filename));

  while ((c = getopt(argc, argv, "f:u:p:d:h")) != EOF) {
    switch (c) {
    case 'u':
      (void)strcpy(username, optarg);
      user = 1;
      noninteractive = 1;
      break;
    case 'p':
      (void)strcpy(password, optarg);
      pword = 1;
      break;
    case 'f':
      (void)strcpy(filename, optarg);
      break;
    case 'd':
      maxdepth = atoi(optarg);
      break;
    case 'h':
      print_version();
      usage();
      break;
    default:
      break;
    }
  }
  if (pword != user) {
    (void)fprintf(stderr, "Must use both -u and -p or neither.\n");
    usage();
  }
  if (maxdepth == 0)
    maxdepth = MAXDEPTH;

  switch (argc - optind) {
  case 0:
    if ((db = (char *)getenv("LINCKSDBDIR")) == (char *)0)
      usage();
    (void)strcpy(dbdir, db);
    break;
  case 1:
    (void)strcpy(dbdir, argv[argc - 1]);
    break;
  default:
    usage();
    break;
  } 
}

/*  */
/**********************************************************************
 * Function: static void usage()
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void usage()
{
  (void)fprintf(stderr, "%s\n%s\n",
			"Usage: dbdump [-h] [-u username -p password]", 
			"       [-f outfile] [-d maxdepth] database-dir");
  exit(1);
}

/*  */
/**********************************************************************
 * Function: static FILE *do_login()
 *
 * Modifications:
 *      <list mods with name and date>
 */
static FILE *do_login()
{
  FILE *f;

  if (noninteractive) {
    if (NOPROMPT_LOGIN(username, password, dbdir) != SUCCESS) {
      (void)fprintf(stderr, "NOPROMPT_LOGIN: login failure\n");
      exit(-1);
    }
  } else {
    if (LOGIN(dbdir) != SUCCESS) {
      (void)fprintf(stderr, "LOGIN: login failure\n");
      exit(-1);
    }
  }

  if (filename[0] == '\0') {
    (void)fprintf(stderr, "No output file.  Writing to stdout.\n");
    if ((f = fdopen(fileno(stdout), "w")) == (FILE *) 0) {
      perror("Failed to open stdout");
      exit(-1);
    }
  } else {
    if ((f = fopen(filename, "wr")) == (FILE *) 0) {
      (void)fprintf(stderr, "Failed to open file %s: ", filename);
      perror("");
      exit(-1);
    }
  }
  return f;
}

#define VERSION "1.1"
#define DATE "1993-09-27"
#define VERSION_INFO "\n"

/*  */
/**********************************************************************
 * Function: static void print_version()
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void print_version()
{
  (void)fprintf(stdout, "\tdbdump version %s (%s)\n%s", VERSION, DATE,
                VERSION_INFO);

}
