/* 
 * Linkoping Intelligent Communication of Knowledge System (LINCKS)
 *      Copyright (C) 1993, Lin Padgham, Ralph Rnnquist
 *       Department of Computer and Information Sciences
 *		University of Linkoping, Sweden
 *		    581 83 Linkoping, Sweden
 *		       lincks@ida.liu.se
 *
 * These collective LINCKS programs are free software; you can 
 * redistribute them and/or modify them under the terms of the GNU
 * General Public License as published by the Free Software Foundation,
 * version 2 of the License.
 *
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the programs; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * MODULE:		dbstat.c
 *
 * SCCSINFO:		@(#)dbstat.c	1.2 11/4/93
 *
 * AUTHOR:		Ralph R|nnquist, 1989-11-20
 *
 * DESCRIPTION:
 *	This file contains the dbstat program, which queries for presence
 *	of a database monitor without actually connecting to it.
 *
 *	Low-level liblincks routine to establish a conenction between
 *	a workspace (application) and database server.
 *
 * MODIFICATIONS:
 *	<list mods with name and date>
 */
/*********************************************************************
 * INCLUDES:
 *********************************************************************/
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>		/* for MAXHOSTNAMELEN */
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <rpc/rpc.h>
#include "xconfig.h"
#include "libshared.h"
#include "systypes.h"

/*********************************************************************
 * EXTERNALLY-CALLABLE ROUTINES FOUND IN THIS MODULE:
 *********************************************************************/
void main( /* int argc, char *(*argv)[] */ );

/*********************************************************************
 * EXTERNALLY-AVAILABLE	DATA FOUND IN THIS MODULE:
 *********************************************************************/
/* none */

/*********************************************************************
 * EXTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
/* none */

/*********************************************************************
 * EXTERNAL DATA STRUCTURES USED BY THIS MODULE:
 *********************************************************************/
/* none */

/*********************************************************************
 * LOCAL DEFINES, STRUCTS, TYPEDEFS, ETC.:
 *********************************************************************/
#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 64
#endif
#define Info(text,value) (void)fprintf(stderr,"\t%s: %s\n",text,value)

/*********************************************************************
 * INTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
static int getdbstatus( /* char *host,int *sock */ );

/*********************************************************************
 * INTERNAL (STATIC) DATA: 
 *********************************************************************/
/* none */

/*  */
/************************************************************************
 * Function: static int getdbstatus(char *host,int *sock)
 *
 * Establishes socket connection to a monitor for the present database
 * and reports on failures etc. Returns 0 if something failed, otherwise 1.
 *
 * This function is essentially a copy of LL_FindMonitorHost in
 * file autostart.c, but specialised for this particular task.
 *
 * Modifications:
 *	<list mods here with name and date>
 *
 */
static int getdbstatus(host, sock)
  char *host;
  int *sock;
{
  static struct sockaddr_in server;
  static struct sockaddr *s = (struct sockaddr *)&server;
  static int opt = SO_DEBUG;
  static struct hostent *hp;
  int st;

  /* Present database configuration information.  */
  (void)fprintf(stderr, "Configuration of database %s is\n", DBDIR);
  Info("Internet port number", LL_GetConfig("TCPIPNO"));
  Info("Host for servers", HOST);
  Info("Database owner", OWNER);
  Info("Server program directory", BINARIES);
  Info("Logging directory", LOGDIR);

  /* Use HOST from run-time configuration.  */
  (void)strcpy(host, HOST);

  /* Retrieve host address from system database.  */
  if ((hp = gethostbyname(host)) == (struct hostent *)0) {
    (void)fprintf(stderr, "Host %s has no net address.\n", host);
    return 0;
  }

  /* Run-time configuration. */
#ifdef SVR42
  memset((char*)&server, 0, sizeof(server));
  memcpy((char*)&server.sin_addr, hp->h_addr, hp->h_length);
#else
  server.sin_addr.s_addr = (*((u_long *) hp->h_addr_list[0])); 
#endif
  server.sin_port = htons(((int)TCPIPNO));	
  server.sin_family = hp->h_addrtype;	/* Always */

  /* Tickle the monitor process to check its existence.  */
  if ((st=callrpc(host, ((u_long) MONITORPROGNO), ((u_long) TCPIPNO), 
		  (u_long) 0, xdr_void, (char *)0, xdr_void, (char *)0))) {
    (void)clnt_perrno(st);
    (void)fprintf(stderr, "The monitor is not running.\n");
    return 0;
  }
  (void)fprintf(stderr, "\tThe monitor is running.\n");

  /* Create a socket for communication and export socket handle.  */
  if ((*sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
    perror("Couldn't create a socket. Very serious problem! ");
    return 0;
  }

  /* Tie socket to server address. (Actually the netserver process).  */
  if (connect(*sock, s, sizeof(server)) != 0) {
    (void)perror("netserv doesn't respond.");
    (void)close(*sock);
    return 0;
  }
  (void)fprintf(stderr, "\tnetserv responds.\n");

#ifdef IP_OPTIONS 
  /*
   * Set socket options to avoid transfer delays. This forces every
   * transfer packet to be available as soon as it comes rather than
   * having the network to wait 100ms for additional bytes.
   */
  if (setsockopt(*sock, IPPROTO_TCP, TCP_NODELAY, &opt, sizeof(int)) < 0) {
    perror("Couldn't configure the socket. Very serious problem! ");
    (void)close(*sock);
    return 0;
  }
#else
  (void) fprintf( stderr, "Warning - TCP_NODELAY not supported.\n");
#endif /* IP_OPTIONS */

  return 1;
}

/*  */
/**********************************************************************
 * Function: void main(int argc, char *(*argv)[])
 * 
 * Modifications:
 *      <list mods with name and date>
 */
void main(argc, argv)
  int argc;
  char *(*argv)[];
{
  static char host[MAXHOSTNAMELEN];
  static int socket;
  char *db = NULL;

  if (argc < 2) {
    if ((db = (char *)getenv("LINCKSDBDIR")) == (char *)0) { 
      (void)fprintf(stderr, "Usage: %s database-dir\n", (*argv)[0]);
      exit(1);
    }
  } else {
    db = (*argv)[1];
  }
  if (configuration(db) == 0) {
    (void)fprintf(stderr, "Couldn't configure %s\n", db);
    exit(1);
  }
  if (getdbstatus(host, &socket) == 1) {
    (void)fprintf(stderr,
		  "The database is available for workspace connection.\n");
    exit(0);
  }
  exit(1);
}
